from unittest import TestCase
from unittest.mock import patch

import pandas as pd

from magnumapi.tool_adapters.roxie.RoxieToolAdapter import RoxieToolAdapter
from tests.resource_files import create_resources_file_path


class TestRoxieToolAdapter(TestCase):
    @patch("plotly.graph_objects.Figure.show")
    def test_parse_roxie_xml_11T(self, mock_show=None):
        # arrange
        roxie_data_xml_path = create_resources_file_path('resources/geometry/roxie/11T/reference/roxieData11T.xml')
        strand_data = RoxieToolAdapter.parse_roxie_xml(roxie_data_xml_path)

        # act
        RoxieToolAdapter.plotly_results(strand_data, column='|B|')

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    def test_parse_roxie_xml_16T_formatted(self):
        # arrange
        roxie_data_xml_path = create_resources_file_path('resources/geometry/roxie/16T/reference/roxieData16T.xml')

        # assert
        with self.assertRaises(IndexError):
            RoxieToolAdapter.parse_roxie_xml(roxie_data_xml_path)

    @patch("plotly.graph_objects.Figure.show")
    def test_parse_roxie_xml_16T_raw(self, mock_show=None):
        # arrange
        roxie_data_xml_path = create_resources_file_path('resources/geometry/roxie/16T/roxieData.xml')
        roxie_data_formatted_xml_path = create_resources_file_path(
            'resources/geometry/roxie/16T/roxieData_formatted.xml')

        # act
        RoxieToolAdapter.correct_xml_file(roxie_data_xml_path, roxie_data_formatted_xml_path)

        strand_data = RoxieToolAdapter.parse_roxie_xml(roxie_data_formatted_xml_path)
        RoxieToolAdapter.plotly_results(strand_data, column='|B|', xlim=(0, 85), ylim=(0, 85))

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("plotly.graph_objects.Figure.show")
    def test_parse_roxie_xml_16T_two_variable_raw(self, mock_show=None):
        # arrange
        roxie_data_xml_path = create_resources_file_path('resources/tool_adapters/roxie/roxieDataTwoVariables.xml')
        roxie_data_formatted_xml_path = create_resources_file_path(
            'resources/tool_adapters/roxie/roxieDataTwoVariables_formatted.xml')

        # act
        RoxieToolAdapter.correct_xml_file(roxie_data_xml_path, roxie_data_formatted_xml_path)

        strand_data_df = RoxieToolAdapter.parse_roxie_xml(roxie_data_formatted_xml_path)
        RoxieToolAdapter.plotly_results(strand_data_df, column='|B|', xlim=(0, 85), ylim=(0, 85))
        RoxieToolAdapter.plotly_results(strand_data_df, column='B3', xlim=(0, 85), ylim=(0, 85))

        roxie_data_xml_path = create_resources_file_path('resources/tool_adapters/roxie/strand_data_ref.csv')
        strand_data_ref_df = pd.read_csv(roxie_data_xml_path, index_col=0)

        pd.testing.assert_frame_equal(strand_data_df, strand_data_ref_df)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    def test_read_figures_of_merit_table_from_output(self):
        # arrange
        roxie_output = create_resources_file_path('resources/tool_adapters/roxie/input.output')

        # act
        fom_df = RoxieToolAdapter.read_figures_of_merit_table_from_output(roxie_output)

        # assert
        fom_ref_df = pd.DataFrame([{'S1': 3, 'S2': 1, 'OPER': 'MINABS', 'OBJECTIVE.1': 89.57, 'WEIGHTED OBJ': 8.957},
                                   {'S1': 5, 'S2': 1, 'OPER': 'MINABS', 'OBJECTIVE.1': 8.257, 'WEIGHTED OBJ': 0.8257},
                                   {'S1': 7, 'S2': 1, 'OPER': 'MINABS', 'OBJECTIVE.1': 12.96, 'WEIGHTED OBJ': 1.296},
                                   {'S1': 0, 'S2': 0, 'OPER': '=2', 'OBJECTIVE.1': 65.14, 'WEIGHTED OBJ': 2514.0},
                                   {'S1': 1, 'S2': 1, 'OPER': '=2', 'OBJECTIVE.1': -4.539, 'WEIGHTED OBJ': 218.9}],
                                  index=['B', 'B', 'B', 'MARGMI', 'BIGB'])
        fom_ref_df.index.names = ['OBJECTIVE']


        pd.testing.assert_frame_equal(fom_ref_df, fom_df)

    def test_convert_figures_of_merit_to_dict(self):
        # arrange
        roxie_output = create_resources_file_path('resources/tool_adapters/roxie/input.output')

        # act
        fom_df = RoxieToolAdapter.read_figures_of_merit_table_from_output(roxie_output)
        fom_dct = RoxieToolAdapter.convert_figures_of_merit_to_dict(fom_df)

        # assert
        fom_dct_ref = {'B_3_1': 89.57, 'B_5_1': 8.257, 'B_7_1': 12.96, 'MARGMI_0_0': 65.14, 'BIGB_1_1': -4.539}

        assert fom_dct == fom_dct_ref



import math
from unittest import TestCase

import dataclasses

from magnumapi.geometry.primitives.Arc import Arc
from magnumapi.geometry.primitives.Point import Point


class TestArc(TestCase):
    def test_calculate_theta_in_rad(self):
        # arrange
        p_start = Point.of_cartesian(10, 0)
        p_end = Point.of_cartesian(0, 10)

        # act
        # assert
        self.assertAlmostEqual(math.radians(90), Arc.calculate_theta_in_rad(p_start, p_end), places=6)

    def test_get_start_angle_in_rad(self):
        # arrange
        p_start = Point.of_cartesian(10, 0)
        p_end = Point.of_cartesian(0, 10)

        # act
        arc = Arc.of_end_points_center(p_start, p_end)
        # assert
        self.assertAlmostEqual(math.radians(0), arc.get_start_angle_in_rad(), places=6)

    def test_get_start_angle(self):
        # arrange
        p_start = Point.of_cartesian(10, 0)
        p_end = Point.of_cartesian(0, 10)

        # act
        arc = Arc.of_end_points_center(p_start, p_end)
        # assert
        self.assertAlmostEqual(0, arc.get_start_angle(), places=6)

    def test_get_end_angle_in_rad(self):
        # arrange
        p_start = Point.of_cartesian(10, 0)
        p_end = Point.of_cartesian(0, 10)

        # act
        arc = Arc.of_end_points_center(p_start, p_end)
        # assert
        self.assertAlmostEqual(math.radians(90), arc.get_end_angle_in_rad(), places=6)

    def test_get_end_angle(self):
        # arrange
        p_start = Point.of_cartesian(10, 0)
        p_end = Point.of_cartesian(0, 10)

        # act
        arc = Arc.of_end_points_center(p_start, p_end)
        # assert
        self.assertAlmostEqual(90, arc.get_end_angle(), places=6)

    def test_get_radius(self):
        # arrange
        p_start = Point.of_cartesian(10, 0)
        p_end = Point.of_cartesian(0, 10)

        # act
        arc = Arc.of_end_points_center(p_start, p_end)
        # assert
        self.assertAlmostEqual(10, arc.get_radius(), places=6)

    def test_describe_ellipse_arc_open(self):
        # arrange
        # act
        arc_str = Arc.describe_ellipse_arc(x_center=0.0, y_center=0.0, a=1.0, b=1.0, start_angle=0.0,
                                           end_angle=2 * math.pi, N=3,
                                           closed=False)

        # assert
        arc_str_ref = 'M 1.0, 0.0L-1.0, 1.2246467991473532e-16L1.0, -2.4492935982947064e-16'
        self.assertEqual(arc_str_ref, arc_str)

    def test_describe_ellipse_arc_closed(self):
        # arrange
        # act
        arc_str = Arc.describe_ellipse_arc(x_center=0.0, y_center=0.0, a=1.0, b=1.0, start_angle=0.0,
                                           end_angle=2 * math.pi, N=3,
                                           closed=True)

        # assert
        arc_str_ref = 'M 1.0, 0.0L-1.0, 1.2246467991473532e-16L1.0, -2.4492935982947064e-16 Z'
        self.assertEqual(arc_str_ref, arc_str)

    def test_immutability_p_start(self):
        # arrange
        p_start = Point.of_cartesian(10, 0)
        p_end = Point.of_cartesian(0, 10)

        # act
        arc = Arc.of_end_points_center(p_start, p_end)

        # act
        with self.assertRaises(dataclasses.FrozenInstanceError):
            exec('arc.p_start = p_end')

    def test_immutability_p_center(self):
        # arrange
        p_start = Point.of_cartesian(10, 0)
        p_end = Point.of_cartesian(0, 10)

        # act
        arc = Arc.of_end_points_center(p_start, p_end)

        # act
        with self.assertRaises(dataclasses.FrozenInstanceError):
            exec('arc.p_center = p_end')

    def test_immutability_theta(self):
        # arrange
        p_start = Point.of_cartesian(10, 0)
        p_end = Point.of_cartesian(0, 10)

        # act
        arc = Arc.of_end_points_center(p_start, p_end)

        # act
        with self.assertRaises(dataclasses.FrozenInstanceError):
            exec('arc.theta = 90')


from unittest import TestCase
from unittest.mock import patch

from magnumapi.geometry.blocks.Block import Block


class TestBlock(TestCase):

    @patch.multiple(Block, __abstractmethods__=set())
    def test_plot_block(self):
        # arrange
        block = Block(cable_def=None,
                      insul_def=None,
                      strand_def=None,
                      conductor_def=None)
        # act
        # assert
        with self.assertRaises(NotImplementedError) as context:
            block.plot_block(None)

        self.assertTrue('This method is not implemented for this class' in str(context.exception))

    @patch.multiple(Block, __abstractmethods__=set())
    def test_build_block(self):
        # arrange
        block = Block(cable_def=None,
                      insul_def=None,
                      strand_def=None,
                      conductor_def=None)
        # act
        # assert
        with self.assertRaises(NotImplementedError) as context:
            block.build_block()

        self.assertTrue('This method is not implemented for this class' in str(context.exception))

    @patch.multiple(Block, __abstractmethods__=set())
    def test_to_roxie_df(self):
        # arrange
        block = Block(cable_def=None,
                      insul_def=None,
                      strand_def=None,
                      conductor_def=None)
        # act
        # assert
        with self.assertRaises(NotImplementedError) as context:
            block.to_block_df()

        self.assertTrue('This method is not implemented for this class' in str(context.exception))



import math
from unittest import TestCase
from unittest.mock import patch

import dataclasses
import matplotlib.pyplot as plt

from magnumapi.geometry.primitives.Point import Point


class TestPoint(TestCase):
    def test_of_cartesian(self):
        # arrange
        x = 1
        y = 5

        # act
        p = Point.of_cartesian(x, y)

        # assert
        self.assertEqual(x, p.x)
        self.assertEqual(y, p.y)

    def test_of_polar(self):
        # arrange
        r = 1
        phi = 60

        # act
        p = Point.of_polar(r, phi)

        # assert
        self.assertAlmostEqual(r, p.get_r(), places=6)
        self.assertAlmostEqual(phi, p.get_phi(), places=6)

    def test_copy(self):
        # arrange
        x = 1
        y = 5
        p = Point.of_cartesian(x, y)

        # act
        p_copy = p.copy()

        # assert
        self.assertNotEqual(id(p), id(p_copy))
        self.assertEqual(id(p), id(p))
        self.assertEqual(id(p_copy), id(p_copy))

    def test_x(self):
        # arrange
        x = 1
        y = 5
        p = Point.of_cartesian(x, y)

        # act
        x_act = p.x

        # assert
        self.assertEqual(x, x_act)

    def test_get_phi(self):
        # arrange
        r = 1
        phi = 60

        # act
        p = Point.of_polar(r, phi)

        # assert
        self.assertAlmostEqual(phi, p.get_phi(), places=6)

    def test_get_phi_in_rad(self):
        # arrange
        r = 1
        phi = 60

        # act
        p = Point.of_polar(r, phi)

        # assert
        self.assertAlmostEqual(2 * math.pi * phi / 360, p.get_phi_in_rad(), places=6)

    def test_get_r(self):
        # arrange
        r = 1
        phi = 60

        # act
        p = Point.of_polar(r, phi)

        # assert
        self.assertAlmostEqual(r, p.get_r(), places=6)

    def test_rotate(self):
        # arrange
        r = 1
        phi = 60
        p = Point.of_polar(r, phi)
        alpha = 15

        # act
        p_rotated = p.rotate(alpha)

        # assert
        self.assertAlmostEqual(phi + alpha, p_rotated.get_phi(), places=6)
        self.assertAlmostEqual(r, p_rotated.get_r(), places=6)

    def test_translate(self):
        # arrange
        x = 1
        y = 5
        p = Point.of_cartesian(x, y)

        x_trans = 2
        y_trans = 2
        p_trans = Point.of_cartesian(x_trans, y_trans)

        # act
        p_new = p.translate(p_trans)

        # assert
        self.assertEqual(x + x_trans, p_new.x)
        self.assertEqual(y + y_trans, p_new.y)

    @patch("matplotlib.pyplot.show")
    def test_plot(self, mock_show=None):
        # arrange
        fig, ax = plt.subplots(figsize=(15, 15))
        x = 1
        y = 5
        p = Point.of_cartesian(x, y)

        # act
        p.plot(ax)
        plt.show()

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    def test_det(self):
        # arrange
        x = 1
        y = 5
        p = Point.of_cartesian(x, y)

        # act
        det = p.det()

        # assert
        self.assertAlmostEqual(math.sqrt(x ** 2 + y ** 2), det, places=6)

    def test_dot(self):
        # arrange
        x = 1
        y = 5
        p = Point.of_cartesian(x, y)

        x_trans = 2
        y_trans = 2
        p2 = Point.of_cartesian(x_trans, y_trans)

        # act
        dot = p.dot(p2)

        #
        self.assertEqual(12, dot)

    def test_unit(self):
        # arrange
        x = 3
        y = 4
        p = Point.of_cartesian(x, y)

        # act
        unit = p.unit()

        # assert
        self.assertEqual(x / 5, unit.x)
        self.assertEqual(y / 5, unit.y)

    def test_unit_zero_length(self):
        # arrange
        x = 0
        y = 0
        p = Point.of_cartesian(x, y)

        # act
        unit = p.unit()

        # assert
        self.assertEqual(0, unit.x)
        self.assertEqual(0, unit.y)

    def test_str_int_int(self):
        # arrange
        x = 3
        y = 4
        p = Point.of_cartesian(x, y)

        # act
        # assert
        self.assertEqual('(3, 4)', str(p))

    def test_str_int_float(self):
        # arrange
        x = 3
        y = 4.0
        p = Point.of_cartesian(x, y)

        # act
        # assert
        self.assertEqual('(3, 4.000000)', str(p))

    def test_str_float_int(self):
        # arrange
        x = 3.0
        y = 4
        p = Point.of_cartesian(x, y)

        # act
        # assert
        self.assertEqual('(3.000000, 4)', str(p))

    def test_str_float_float(self):
        # arrange
        x = 3.0
        y = 4.0
        p = Point.of_cartesian(x, y)

        # act
        # assert
        self.assertEqual('(3.000000, 4.000000)', str(p))

    def test_repr(self):
        # arrange
        x = 3
        y = 4
        p = Point.of_cartesian(x, y)

        # act
        # assert
        self.assertEqual('Point(3.000000, 4.000000)', repr(p))

    def test_equality(self):
        # arrange
        x = 3
        y = 4
        p1 = Point.of_cartesian(x, y)
        p2 = Point.of_cartesian(x, y)

        # act
        # assert
        self.assertTrue(p1 == p2)

    def test_non_equality(self):
        # arrange
        x = 3
        y = 4
        p1 = Point.of_cartesian(x, y)
        x = 3
        y = 5
        p2 = Point.of_cartesian(x, y)

        # act
        # assert
        self.assertFalse(p1 == p2)

    def test_immutability_x(self):
        # arrange
        x = 3
        y = 4
        p1 = Point.of_cartesian(x, y)

        # act
        with self.assertRaises(dataclasses.FrozenInstanceError):
            exec('p1.x = 5')

    def test_immutability_y(self):
        # arrange
        x = 3
        y = 4
        p1 = Point.of_cartesian(x, y)

        # act
        with self.assertRaises(dataclasses.FrozenInstanceError):
            exec('p1.y = 5')



import os

import pytest

from magnumapi.tool_adapters.ansys.AnsysToolAdapter import DockerMapdlAnsysToolAdapter

config = {"input_folder_rel_dir": "input\\Mechanic_Plate",
          "input_file": "15T_mech.inp",
          "output_file": "vallone.out",
          "model_file": "Model.inp",
          "rst_file": "15T_2d_mech.rst",
          "single_upload_files": [],
          "reupload_files": [],
          "ip": "127.0.0.1",
          "port": 50052}


def test_constructor():
    # arrange

    # act
    ansys = DockerMapdlAnsysToolAdapter(**config)

    # assert
    assert os.getcwd() == ansys.root_dir
    assert "input\\Mechanic_Plate" == ansys.input_folder_rel_dir
    assert "15T_mech.inp" == ansys.input_file
    assert "vallone.out" == ansys.output_file
    assert "Model.inp" == ansys.model_file
    assert "15T_2d_mech.rst" == ansys.rst_file
    assert [] == ansys.single_upload_files
    assert [] == ansys.reupload_files
    assert "127.0.0.1" == ansys.ip
    assert 50052 == ansys.port



from unittest import TestCase

from magnumapi.optimization.ParallelOptimizationConfig import ParallelOptimizationConfig
from tests.resource_files import create_resources_file_path


class TestParallelOptimizationConfig(TestCase):
    def setUp(self) -> None:
        json_path = create_resources_file_path('resources/optimization/parallel_config.json')
        self.config = ParallelOptimizationConfig.initialize_config(json_path)

    def test_initialize_config(self):
        # arrange
        # act
        # assert
        self.assertEqual("c:\\gitlab\\magnum-nb", self.config.input_folder)
        self.assertEqual("c:\\magnum", self.config.output_folder)
        self.assertEqual('optimization_parallel\\GeneticOptimization.csv', self.config.logger_rel_path)

        self.assertEqual(100, self.config.n_gen)
        self.assertEqual(20, self.config.n_pop)
        self.assertAlmostEqual(0.9, self.config.r_cross, places=1)
        self.assertAlmostEqual(0.01, self.config.r_mut, places=2)

        self.assertEqual('b3', self.config.objectives[0].objective)
        self.assertAlmostEqual(0.1, self.config.objectives[0].weight, places=1)
        self.assertAlmostEqual(0.0, self.config.objectives[0].constraint, places=1)

        self.assertEqual('b5', self.config.objectives[1].objective)
        self.assertAlmostEqual(0.1, self.config.objectives[1].weight, places=1)
        self.assertAlmostEqual(0.0, self.config.objectives[1].constraint, places=1)

        self.assertEqual('margmi', self.config.objectives[2].objective)
        self.assertAlmostEqual(1, self.config.objectives[2].weight, places=0)
        self.assertAlmostEqual(0.85, self.config.objectives[2].constraint, places=2)

        self.assertEqual('seqv', self.config.objectives[3].objective)
        self.assertAlmostEqual(0.001, self.config.objectives[3].weight, places=3)
        self.assertAlmostEqual(0.0, self.config.objectives[3].constraint, places=1)

        self.assertEqual('geometry_magnetic_for_parallel', self.config.notebooks[0].notebook_folder)
        self.assertEqual('Geometry_ROXIE.ipynb', self.config.notebooks[0].notebook_name)
        self.assertEqual({"index": "index"}, self.config.notebooks[0].input_parameters)
        self.assertEqual(['b3', 'b5', 'bigb', 'margmi'], self.config.notebooks[0].output_parameters)
        self.assertEqual({}, self.config.notebooks[0].input_artefacts)
        self.assertEqual([], self.config.notebooks[0].output_artefacts)

        self.assertEqual('C:\\ansys', self.config.ansys.run_dir)
        self.assertEqual(10, self.config.ansys.n_parallel_runners)
        self.assertEqual(1, self.config.ansys.n_proc)
        self.assertEqual('mechanical_parallel', self.config.ansys.root_dir)
        self.assertEqual("", self.config.ansys.exec_file)

    def test_get_weight(self):
        # arrange
        objective = 'b3'

        # act
        weight = self.config.get_weight(objective)

        # assert
        self.assertAlmostEqual(0.1, weight)

    def test_get_weight_error(self):
        # arrange
        objective = 'b7'

        # act
        with self.assertRaises(KeyError) as context:
            self.config.get_weight(objective)

        # assert
        self.assertTrue('Objective name b7 not present in objective configs.' in str(context.exception))

    def test_get_constraint(self):
        # arrange
        objective = 'b3'

        # act
        constraint = self.config.get_constraint(objective)

        # assert
        self.assertAlmostEqual(0, constraint)

    def test_get_constraint_error(self):
        # arrange
        objective = 'b7'

        # act
        with self.assertRaises(KeyError) as context:
            self.config.get_constraint(objective)

        # assert
        self.assertTrue('Objective name b7 not present in objective configs.' in str(context.exception))



from unittest import TestCase
from unittest.mock import patch

import matplotlib.pyplot as plt

from magnumapi.geometry.GeometryChange import GeometryChange
from magnumapi.geometry.primitives.Area import Area
from magnumapi.geometry.blocks.CosThetaBlock import RelativeCosThetaBlock, CosThetaBlock
from magnumapi.geometry.primitives.Line import Line
from magnumapi.geometry.primitives.Point import Point
from magnumapi.geometry.definitions.RelativeCosThetaBlockDefinition import RelativeCosThetaBlockDefinition
from magnumapi.cadata.CableDatabase import CableDatabase
from tests.resource_files import create_resources_file_path


class TestCosThetaBlock(TestCase):
    cadata_file_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
    cadata = CableDatabase.read_cadata(cadata_file_path)

    # BLOCK 1
    #     1     1     1            0          0              0        13500     16TIL6B   2 20     0            0
    #    no  type   nco       radius          phi        alpha      current    condname  n1 n2  imag         turn

    #  INSUL 1
    #   1 BARE                 0            0 'BARE                    '
    #  No Name            Radial       Azimut  Comment
    #
    #  CABLE 1
    #   1 16TILG6B         21.84       2.2595       2.2595    38          100            3 'EUROCIRCOL 26 STR             '
    #  No Name            height      width_i      width_o    ns      transp.        degrd  Comment
    #
    # CONDUCTOR 1
    #   1 16TIL6B       1 16TILG6B     STR11     NB3SN2    BARE    NONE      NB3SN                1.9 'eurocircol                    '
    #  No Name       Type CableGeom.  Strand    Filament  Insul     Trans     QuenchMat.           T_o  Comment
    @patch("matplotlib.pyplot.show")
    def test_get_first_isosceles_trapezium_rect_no_insulation_no_radius_no_phi_no_alpha(self, mock_show=None):
        block_rel_def = RelativeCosThetaBlockDefinition(radius=0, phi_r=0.0, alpha_r=0.0, nco=1, condname='16TIL6B',
                                                        no=1, current=13500, type=1, n1=2, n2=20, imag=0, turn=0)
        block = RelativeCosThetaBlock(block_def=block_rel_def,
                                      cable_def=self.cadata.get_cable_definition('16TIL6B'),
                                      insul_def=self.cadata.get_insul_definition('16TIL6B'),
                                      strand_def=self.cadata.get_strand_definition('16TIL6B'),
                                      conductor_def=self.cadata.get_conductor_definition('16TIL6B'))
        p_ref = Point.of_polar(block.block_def.radius, block.get_phi())
        area_ins = block.get_insulated_isosceles_trapezium(p_ref=p_ref,
                                                           alpha_ref=block.get_alpha())
        area_shift = area_ins.copy()

        if block.is_area_intersecting_or_within_radius(area_ins):
            area_shift = area_ins.translate(Point.of_polar(2 * block.cable_def.thickness_i, block.get_phi()))
            shift = Area.find_trapezoid_shift_to_intercept_with_radius(block.block_def.radius, area_shift)
            area_shift = area_shift.translate(area_ins.get_line(0).unit() * shift)

        fig, ax = plt.subplots(figsize=(15, 15))
        ax.set_aspect('equal', 'box')
        ax.grid(True)
        ax.set_xlim(0, 75)
        ax.set_ylim(0, 75)

        circle = plt.Circle((0, 0), block.block_def.radius, color='r', fill=False)
        ax.add_patch(circle)
        area_ins.plot(ax)
        area_shift.plot(ax)
        plt.show()

        # assert positions - insulated
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 0.0000
        # y(1) = 2.2595
        # x(2) = 0.0000
        # y(2) = 0.0000
        # x(3) = 21.8400
        # y(3) = 0.0000
        # x(4) = 21.8400
        # y(4) = 2.2595
        self.assertAlmostEqual(0.0000, round(area_shift.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(2.2595, round(area_shift.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_shift.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_shift.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(21.8400, round(area_shift.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_shift.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(21.8400, round(area_shift.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(2.2595, round(area_shift.get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0 (1-4) = NaN
        # Alph0 (1-4) = 0.0000
        # Phi0 (2-3) = 90.0000
        # Alph0 (2-3) = 0.0000

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 0.0000
        # y(1) = 2.2595
        # x(2) = 0.0000
        # y(2) = 0.0000
        # x(3) = 21.8400
        # y(3) = 0.0000
        # x(4) = 21.8400
        # y(4) = 2.2595

        area_bare = block.get_bare_isosceles_trapezium(p_ref=area_shift.get_line(0).p1,
                                                       alpha_ref=Line.calculate_relative_alpha_angle(
                                                           area_shift.get_line(0)))

        self.assertAlmostEqual(0.0000, round(area_bare.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(2.2595, round(area_bare.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_bare.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_bare.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(21.8400, round(area_bare.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_bare.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(21.8400, round(area_bare.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(2.2595, round(area_bare.get_line(2).p1.y, 4), places=4)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    # BLOCK 1
    #     1     1     1            0          0              0        13500     16TILG6   2  20     0            0
    #    no  type   nco       radius          phi        alpha      current    condname  n1 n2  imag         turn

    #  INSUL 1
    #   1 16TINS2           0.15         0.15 'eurocircol                    '
    #  No Name            Radial       Azimut  Comment
    #
    #  CABLE 1
    #   1 16TILG6B         21.84       2.2595       2.2595    38          100            3 'EUROCIRCOL 26 STR             '
    #  No Name            height      width_i      width_o    ns      transp.        degrd  Comment
    #
    # CONDUCTOR 1
    #   1 16TIL6C       1 16TILG6B     STR11     NB3SN2   16TINS2    NONE      NB3SN                1.9 'eurocircol                    '
    #  No Name       Type CableGeom.  Strand    Filament  Insul     Trans     QuenchMat.           T_o  Comment
    @patch("matplotlib.pyplot.show")
    def test_get_first_isosceles_trapezium_rect_with_insulation_no_radius_no_phi_no_alpha(self, mock_show=None):
        block_rel_def = RelativeCosThetaBlockDefinition(radius=0, phi_r=0.0, alpha_r=0.0, nco=1, condname='16TIL6B',
                                                        no=1, current=13500, type=1, n1=2, n2=20, imag=0, turn=0)
        block = RelativeCosThetaBlock(block_def=block_rel_def,
                                      cable_def=self.cadata.get_cable_definition('16TIL6C'),
                                      insul_def=self.cadata.get_insul_definition('16TIL6C'),
                                      strand_def=self.cadata.get_strand_definition('16TIL6B'),
                                      conductor_def=self.cadata.get_conductor_definition('16TIL6B'))
        p_ref = Point.of_polar(block.block_def.radius, block.block_def.phi_r)
        area_ins = block.get_insulated_isosceles_trapezium(p_ref=p_ref,
                                                           alpha_ref=block.get_alpha())
        area_shift = area_ins.copy()

        fig, ax = plt.subplots(figsize=(15, 15))
        ax.set_aspect('equal', 'box')
        ax.grid(True)
        ax.set_xlim(0, 75)
        ax.set_ylim(0, 75)

        circle = plt.Circle((0, 0), block.block_def.radius, color='r', fill=False)
        ax.add_patch(circle)
        area_ins.plot(ax)
        area_shift.plot(ax)
        plt.show()

        # assert positions - insulated
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 0.0000
        # y(1) = 2.5595
        # x(2) = 0.0000
        # y(2) = 0.0000
        # x(3) = 22.1400
        # y(3) = 0.0000
        # x(4) = 22.1400
        # y(4) = 2.5595
        self.assertAlmostEqual(0.0000, round(area_shift.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(2.5595, round(area_shift.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_shift.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_shift.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(22.1400, round(area_shift.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_shift.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(22.1400, round(area_shift.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(2.5595, round(area_shift.get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0 (1-4) = NaN
        # Alph0 (1-4) = 0.0000
        # Phi0 (2-3) = 90.0000
        # Alph0 (2-3) = 0.0000

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 0.1500
        # y(1) = 2.5595
        # x(2) = 0.1500
        # y(2) = 0.1500
        # x(3) = 22.1400
        # y(3) = 0.0000
        # x(4) = 22.1400
        # y(4) = 2.5595

        area_bare = block.get_bare_isosceles_trapezium(p_ref=area_shift.get_line(0).p1,
                                                       alpha_ref=Line.calculate_relative_alpha_angle(
                                                           area_shift.get_line(0)))

        self.assertAlmostEqual(0.1500, round(area_bare.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(2.4095, round(area_bare.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(0.1500, round(area_bare.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.1500, round(area_bare.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(21.9900, round(area_bare.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(0.1500, round(area_bare.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(21.9900, round(area_bare.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(2.4095, round(area_bare.get_line(2).p1.y, 4), places=4)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    # BLOCK 1
    #     1     1     1           25          0              0        13500     16TILG6   2  20     0            0
    #    no  type   nco       radius          phi        alpha      current    condname  n1 n2  imag         turn

    #  INSUL 1
    #   1 16TINS2           0.15         0.15 'eurocircol                    '
    #  No Name            Radial       Azimut  Comment
    #
    #  CABLE 1
    #   1 16TILG6B         21.84       2.2595       2.2595    38          100            3 'EUROCIRCOL 26 STR             '
    #  No Name            height      width_i      width_o    ns      transp.        degrd  Comment
    #
    # CONDUCTOR 1
    #   1 16TIL6C       1 16TILG6B     STR11     NB3SN2   16TINS2    NONE      NB3SN                1.9 'eurocircol                    '
    #  No Name       Type CableGeom.  Strand    Filament  Insul     Trans     QuenchMat.           T_o  Comment
    @patch("matplotlib.pyplot.show")
    def test_get_first_isosceles_trapezium_rect_with_insulation_with_radius_no_phi_no_alpha(self, mock_show=None):
        block_rel_def = RelativeCosThetaBlockDefinition(radius=25, phi_r=0.0, alpha_r=0.0, nco=1, condname='16TIL6C',
                                                        no=1,
                                                        current=13500, type=1, n1=2, n2=20, imag=0, turn=0)
        block = RelativeCosThetaBlock(block_def=block_rel_def,
                                      cable_def=self.cadata.get_cable_definition('16TIL6C'),
                                      insul_def=self.cadata.get_insul_definition('16TIL6C'),
                                      strand_def=self.cadata.get_strand_definition('16TIL6B'),
                                      conductor_def=self.cadata.get_conductor_definition('16TIL6B'))
        p_ref = Point.of_polar(block.block_def.radius, block.get_phi())
        area_ins = block.get_insulated_isosceles_trapezium(p_ref=p_ref,
                                                           alpha_ref=block.get_alpha())
        area_shift = area_ins.copy()

        if block.is_area_intersecting_or_within_radius(area_ins):
            area_shift = area_ins.translate(Point.of_polar(2 * block.cable_def.thickness_i, block.get_alpha()))
            shift = Area.find_trapezoid_shift_to_intercept_with_radius(block.block_def.radius, area_shift)
            area_shift = area_shift.translate(area_ins.get_line(0).unit() * shift)

        fig, ax = plt.subplots(figsize=(15, 15))
        ax.set_aspect('equal', 'box')
        ax.grid(True)
        ax.set_xlim(0, 75)
        ax.set_ylim(0, 75)

        circle = plt.Circle((0, 0), block.block_def.radius, color='r', fill=False)
        ax.add_patch(circle)
        area_ins.plot(ax)
        area_shift.plot(ax)
        plt.show()

        # assert positions - insulated
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 25.0000
        # y(1) = 2.5595
        # x(2) = 25.0000
        # y(2) = 0.0000
        # x(3) = 47.1400
        # y(3) = 0.0000
        # x(4) = 47.1400
        # y(4) = 2.5595
        self.assertAlmostEqual(25.0000, round(area_shift.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(2.5595, round(area_shift.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(25.0000, round(area_shift.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_shift.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(47.1400, round(area_shift.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_shift.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(47.1400, round(area_shift.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(2.5595, round(area_shift.get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0 (1-4) = 5.8762
        # Alph0 (1-4) = 0.0000
        # Phi0 (2-3) = 0.0000
        # Alph0 (2-3) = 0.0000
        self.assertAlmostEqual(5.8762, Line.calculate_positioning_angle(area_shift.get_line(2), block.block_def.radius),
                               places=4)
        self.assertAlmostEqual(0.0000, Line.calculate_relative_alpha_angle(area_shift.get_line(2)), places=4)
        self.assertAlmostEqual(0.0000, Line.calculate_positioning_angle(area_shift.get_line(0), block.block_def.radius),
                               places=4)
        self.assertAlmostEqual(0.0000, Line.calculate_relative_alpha_angle(area_shift.get_line(0)), places=4)

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 25.1500
        # y(1) = 2.4095
        # x(2) = 25.1500
        # y(2) = 0.1500
        # x(3) = 46.9900
        # y(3) = 0.1500
        # x(4) = 46.9900
        # y(4) = 2.4095

        area_bare = block.get_bare_isosceles_trapezium(p_ref=area_shift.get_line(0).p1,
                                                       alpha_ref=Line.calculate_relative_alpha_angle(
                                                           area_shift.get_line(0)))

        self.assertAlmostEqual(25.1500, round(area_bare.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(2.4095, round(area_bare.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(25.1500, round(area_bare.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.1500, round(area_bare.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(46.9900, round(area_bare.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(0.1500, round(area_bare.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(46.9900, round(area_bare.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(2.4095, round(area_bare.get_line(2).p1.y, 4), places=4)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    # BLOCK 1
    #     1     1     1           25          5              0        13500     16TIL9   2  20     0            0
    #    no  type   nco       radius          phi        alpha      current    condname  n1 n2  imag         turn
    #  INSUL 1
    #   1 16TINS2           0.15         0.15 'eurocircol                    '
    #  No Name            Radial       Azimut  Comment
    #
    #  CABLE 1
    #   1 16TILG6B         21.84       2.2595       2.2595    38          100            3 'EUROCIRCOL 26 STR             '
    #  No Name            height      width_i      width_o    ns      transp.        degrd  Comment
    #
    # CONDUCTOR 1
    #   1 16TIL6C       1 16TILG6B     STR11     NB3SN2   16TINS2    NONE      NB3SN                1.9 'eurocircol                    '
    #  No Name       Type CableGeom.  Strand    Filament  Insul     Trans     QuenchMat.           T_o  Comment
    @patch("matplotlib.pyplot.show")
    def test_get_first_isosceles_trapezium_rect_with_insulation_with_radius_with_phi_no_alpha(self, mock_show=None):
        block_rel_def = RelativeCosThetaBlockDefinition(radius=25, phi_r=5.0, alpha_r=0.0, nco=1, condname='16TIL6C',
                                                        no=1,
                                                        current=13500, type=1, n1=2, n2=20, imag=0, turn=0)
        block = RelativeCosThetaBlock(block_def=block_rel_def,
                                      cable_def=self.cadata.get_cable_definition('16TIL6C'),
                                      insul_def=self.cadata.get_insul_definition('16TIL6C'),
                                      strand_def=self.cadata.get_strand_definition('16TIL6B'),
                                      conductor_def=self.cadata.get_conductor_definition('16TIL6B'))
        p_ref = Point.of_polar(block.block_def.radius, block.block_def.phi_r)
        area_ins = block.get_insulated_isosceles_trapezium(p_ref=p_ref,
                                                           alpha_ref=block.get_alpha())
        area_shift = area_ins.copy()

        if block.is_area_intersecting_or_within_radius(area_ins):
            area_shift = area_ins.translate(Point.of_polar(2 * block.cable_def.thickness_i, block.get_alpha()))
            shift = Area.find_trapezoid_shift_to_intercept_with_radius(block.block_def.radius, area_shift)
            area_shift = area_shift.translate(area_ins.get_line(0).unit() * shift)

        fig, ax = plt.subplots(figsize=(15, 15))
        ax.set_aspect('equal', 'box')
        ax.grid(True)
        ax.set_xlim(0, 75)
        ax.set_ylim(0, 75)

        circle = plt.Circle((0, 0), block.block_def.radius, color='r', fill=False)
        ax.add_patch(circle)
        area_ins.plot(ax)
        area_shift.plot(ax)
        plt.show()

        # assert positions - insulated
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 24.9049
        # y(1) = 4.7384
        # x(2) = 24.9049
        # y(2) = 2.1789
        # x(3) = 47.0449
        # y(3) = 2.1789
        # x(4) = 47.0449
        # y(4) = 4.7384
        self.assertAlmostEqual(24.9049, round(area_shift.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(4.7384, round(area_shift.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(24.9049, round(area_shift.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(2.1789, round(area_shift.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(47.0449, round(area_shift.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(2.1789, round(area_shift.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(47.0449, round(area_shift.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(4.7384, round(area_shift.get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0 (1-4) = 10.9257
        # Alph0 (1-4) = 0.0000
        # Phi0 (2-3) = 5.0000
        # Alph0 (2-3) = 0.0000
        self.assertAlmostEqual(10.9257, Line.calculate_positioning_angle(area_shift.get_line(2),
                                                                         block.block_def.radius), places=4)
        self.assertAlmostEqual(0.0000, Line.calculate_relative_alpha_angle(area_shift.get_line(2)), places=4)
        self.assertAlmostEqual(5.0000, Line.calculate_positioning_angle(area_shift.get_line(0),
                                                                        block.block_def.radius), places=4)
        self.assertAlmostEqual(0.0000, Line.calculate_relative_alpha_angle(area_shift.get_line(0)), places=4)

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 25.0549
        # y(1) = 4.5884
        # x(2) = 25.0549
        # y(2) = 2.3289
        # x(3) = 46.8949
        # y(3) = 2.3289
        # x(4) = 46.8949
        # y(4) = 4.5884

        area_bare = block.get_bare_isosceles_trapezium(p_ref=area_shift.get_line(0).p1,
                                                       alpha_ref=Line.calculate_relative_alpha_angle(
                                                           area_shift.get_line(0)))

        self.assertAlmostEqual(25.0549, round(area_bare.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(4.5884, round(area_bare.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(25.0549, round(area_bare.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(2.3289, round(area_bare.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(46.8949, round(area_bare.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(2.3289, round(area_bare.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(46.8949, round(area_bare.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(4.5884, round(area_bare.get_line(2).p1.y, 4), places=4)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    # BLOCK 1
    #     1     1     1           25          5             10        13500     16TIL9   2  20     0            0
    #    no  type   nco       radius          phi        alpha      current    condname  n1 n2  imag         turn
    #  INSUL 1
    #   1 16TINS2           0.15         0.15 'EUROCIRCOL                    '
    #  No Name            Radial       Azimut  Comment
    #
    #  CABLE 1
    #   1 16TILG6          21.84       2.2595       2.2595    38          100            3 'EUROCIRCOL 26 STR             '
    #  No Name            height      width_i      width_o    ns      transp.        degrd  Comment
    #
    # CONDUCTOR 1
    #   1 16TIL6        1 16TILG6     STR11     NB3SN2    16TINS    NONE      NB3SN                1.9 'eurocircol                    '
    #  No Name       Type CableGeom.  Strand    Filament  Insul     Trans     QuenchMat.           T_o  Comment
    @patch("matplotlib.pyplot.show")
    def test_get_first_isosceles_trapezium_rect_with_insulation_with_radius_with_phi_with_alpha(self, mock_show=None):
        block_rel_def = RelativeCosThetaBlockDefinition(radius=25, phi_r=5.0, alpha_r=10.0, nco=1, condname='16TIL6C',
                                                        no=1, current=13500, type=1, n1=2, n2=20, imag=0, turn=0)
        block = RelativeCosThetaBlock(block_def=block_rel_def,
                                      cable_def=self.cadata.get_cable_definition('16TIL6C'),
                                      insul_def=self.cadata.get_insul_definition('16TIL6C'),
                                      strand_def=self.cadata.get_strand_definition('16TIL6B'),
                                      conductor_def=self.cadata.get_conductor_definition('16TIL6B'))
        p_ref = Point.of_polar(block.block_def.radius, block.get_phi())
        area_ins = block.get_insulated_isosceles_trapezium(p_ref=p_ref,
                                                           alpha_ref=block.get_alpha())
        area_shift = area_ins.copy()

        if block.is_area_intersecting_or_within_radius(area_ins):
            area_shift = area_ins.translate(Point.of_polar(2 * block.cable_def.thickness_i, block.get_alpha()))
            shift = Area.find_trapezoid_shift_to_intercept_with_radius(block.block_def.radius, area_shift, eps=1e-9)
            area_shift = area_shift.translate(area_ins.get_line(0).unit() * shift)

        fig, ax = plt.subplots(figsize=(15, 15))
        ax.set_aspect('equal', 'box')
        ax.grid(True)
        ax.set_xlim(22, 27)
        ax.set_ylim(0, 10)

        circle = plt.Circle((0, 0), block.block_def.radius, color='r', fill=False)
        ax.add_patch(circle)
        area_ins.plot(ax)
        area_shift.plot(ax)
        plt.show()

        # assert positions - insulated
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 24.5541
        # y(1) = 4.7160
        # x(2) = 24.9986
        # y(2) = 2.1954
        # x(3) = 46.8022
        # y(3) = 6.0400
        # x(4) = 46.3577
        # y(4) = 8.5606

        self.assertAlmostEqual(24.5541, round(area_shift.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(4.7160, round(area_shift.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(24.9986, round(area_shift.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(2.1954, round(area_shift.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(46.8022, round(area_shift.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(6.0400, round(area_shift.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(46.3577, round(area_shift.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(8.5606, round(area_shift.get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0 (1-4) = 10.8723
        # Alph0 (1-4) = 10.0000
        # Phi0 (2-3) = 5.0000
        # Alph0 (2-3) = 10.0000
        self.assertAlmostEqual(10.8723, Line.calculate_positioning_angle(area_shift.get_line(2),
                                                                         block.block_def.radius), places=4)
        self.assertAlmostEqual(10.0000, Line.calculate_relative_alpha_angle(area_shift.get_line(2)), places=4)
        self.assertAlmostEqual(5.0000, Line.calculate_positioning_angle(area_shift.get_line(0),
                                                                        block.block_def.radius), places=4)
        self.assertAlmostEqual(10.0000, Line.calculate_relative_alpha_angle(area_shift.get_line(0)), places=4)

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 24.7279
        # y(1) = 4.5944
        # x(2) = 25.1202
        # y(2) = 2.3692
        # x(3) = 46.6284
        # y(3) = 6.1617
        # x(4) = 46.2361
        # y(4) = 8.3868

        area_bare = block.get_bare_isosceles_trapezium(p_ref=area_shift.get_line(0).p1,
                                                       alpha_ref=Line.calculate_relative_alpha_angle(
                                                           area_shift.get_line(0)))

        self.assertAlmostEqual(24.7279, round(area_bare.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(4.5944, round(area_bare.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(25.1202, round(area_bare.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(2.3692, round(area_bare.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(46.6284, round(area_bare.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(6.1617, round(area_bare.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(46.2361, round(area_bare.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(8.3868, round(area_bare.get_line(2).p1.y, 4), places=4)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    # BLOCK 1
    #     1     1     4            0            0            0        13500     16TIL9   2  20     0            0
    #    no  type   nco       radius          phi        alpha      current    condname  n1 n2  imag         turn
    #  INSUL 1
    #   1 BARE                 0            0 'BARE                    '
    #  No Name            Radial       Azimut  Comment
    #
    #  CABLE 1
    #   1 16TOLG78B        21.84       1.8782       2.2595    38          100            3 'EUROCIRCOL 26 STR             '
    #  No Name            height      width_i      width_o    ns      transp.        degrd  Comment
    #
    # CONDUCTOR 1
    #   1 16TIL6D        1 16TOLG78B  STR11     NB3SN2    BARE      NONE      NB3SN                1.9 'eurocircol                    '
    #  No Name       Type CableGeom.  Strand    Filament  Insul     Trans     QuenchMat.           T_o  Comment
    @patch("matplotlib.pyplot.show")
    def test_get_first_isosceles_trapezium_trapz_no_insulation_no_radius_no_phi_no_alpha(self, mock_show=None):
        block_rel_def = RelativeCosThetaBlockDefinition(radius=0, phi_r=0.0, alpha_r=0.0, nco=1, condname='16TIL6D',
                                                        no=1,
                                                        current=13500, type=1, n1=2, n2=20, imag=0, turn=0)
        block = RelativeCosThetaBlock(block_def=block_rel_def,
                                      cable_def=self.cadata.get_cable_definition('16TIL6D'),
                                      insul_def=self.cadata.get_insul_definition('16TIL6D'),
                                      strand_def=self.cadata.get_strand_definition('16TIL6D'),
                                      conductor_def=self.cadata.get_conductor_definition('16TIL6D'))
        p_ref = Point.of_polar(block.block_def.radius, block.get_phi())
        area_ins = block.get_insulated_isosceles_trapezium(p_ref=p_ref,
                                                           alpha_ref=block.get_alpha())
        area_shift = area_ins.copy()

        if block.is_area_intersecting_or_within_radius(area_ins):
            area_shift = area_ins.translate(Point.of_polar(2 * block.cable_def.thickness_i, block.get_alpha()))
            shift = Area.find_trapezoid_shift_to_intercept_with_radius(block.block_def.radius, area_shift)
            area_shift = area_shift.translate(area_ins.get_line(0).unit() * shift)

        fig, ax = plt.subplots(figsize=(15, 15))
        ax.set_aspect('equal', 'box')
        ax.grid(True)
        ax.set_xlim(0, 75)
        ax.set_ylim(0, 75)

        circle = plt.Circle((0, 0), block.block_def.radius, color='r', fill=False)
        ax.add_patch(circle)
        area_ins.plot(ax)
        area_shift.plot(ax)
        plt.show()

        # assert positions - insulated
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = -0.0164
        # y(1) = 1.8781
        # x(2) = 0.0000
        # y(2) = 0.0000
        # x(3) = 41.8408
        # y(3) = 0.0000
        # x(4) = 21.8211
        # y(4) = 2.2594
        self.assertAlmostEqual(-0.0164, round(area_shift.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(1.8781, round(area_shift.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_shift.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_shift.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(21.8408, round(area_shift.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_shift.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(21.8211, round(area_shift.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(2.2594, round(area_shift.get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0 (1-4) = nan
        # Alph0 (1-4) = 1.003
        # Phi0 (2-3) = 0.0000
        # Alph0 (2-3) = 0.0000
        self.assertAlmostEqual(1.0003, Line.calculate_relative_alpha_angle(area_shift.get_line(2)), places=4)

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = -0.0164
        # y(1) = 1.8781
        # x(2) = 0.0000
        # y(2) = 0.0000
        # x(3) = 41.8408
        # y(3) = 0.0000
        # x(4) = 21.8211
        # y(4) = 2.2594

        area_bare = block.get_bare_isosceles_trapezium(p_ref=area_shift.get_line(0).p1,
                                                       alpha_ref=Line.calculate_relative_alpha_angle(
                                                           area_shift.get_line(0)))

        self.assertAlmostEqual(-0.0164, round(area_bare.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(1.8781, round(area_bare.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_bare.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_bare.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(21.8408, round(area_bare.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_bare.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(21.8211, round(area_bare.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(2.2594, round(area_bare.get_line(2).p1.y, 4), places=4)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    # BLOCK 1
    #     1     1     4            0            0            0        13500     16TIL9   2  20     0            0
    #    no  type   nco       radius          phi        alpha      current    condname  n1 n2  imag         turn
    #  INSUL 1
    #   1 16TINS2           0.15         0.15 'eurocircol                    '
    #  No Name            Radial       Azimut  Comment
    #
    #  CABLE 1
    #   1 16TOLG78B        21.84       1.8782       2.2595    38          100            3 'EUROCIRCOL 26 STR             '
    #  No Name            height      width_i      width_o    ns      transp.        degrd  Comment
    #
    # CONDUCTOR 1
    #   1 16TIL6E        1 16TOLG78B  STR11     NB3SN2    16TINS2   NONE      NB3SN                1.9 'eurocircol                    '
    #  No Name       Type CableGeom.  Strand    Filament  Insul     Trans     QuenchMat.           T_o  Comment
    @patch("matplotlib.pyplot.show")
    def test_get_first_isosceles_trapezium_trapz_with_insulation_no_radius_no_phi_no_alpha(self, mock_show=None):
        block_rel_def = RelativeCosThetaBlockDefinition(radius=0, phi_r=0.0, alpha_r=0.0, nco=1, condname='16TIL6E',
                                                        no=1,
                                                        current=13500, type=1, n1=2, n2=20, imag=0, turn=0)
        block = RelativeCosThetaBlock(block_def=block_rel_def,
                                      cable_def=self.cadata.get_cable_definition('16TIL6E'),
                                      insul_def=self.cadata.get_insul_definition('16TIL6E'),
                                      strand_def=self.cadata.get_strand_definition('16TIL6E'),
                                      conductor_def=self.cadata.get_conductor_definition('16TIL6E'))
        p_ref = Point.of_polar(block.block_def.radius, block.get_phi())
        area_ins = block.get_insulated_isosceles_trapezium(p_ref=p_ref,
                                                           alpha_ref=block.get_alpha())
        area_shift = area_ins.copy()

        fig, ax = plt.subplots(figsize=(15, 15))
        ax.set_aspect('equal', 'box')
        ax.grid(True)
        ax.set_xlim(0, 75)
        ax.set_ylim(0, 75)

        circle = plt.Circle((0, 0), block.block_def.radius, color='r', fill=False)
        ax.add_patch(circle)
        area_ins.plot(ax)
        area_shift.plot(ax)
        plt.show()

        # assert positions - insulated
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = -0.0190
        # y(1) = 2.1755
        # x(2) = 0.0000
        # y(2) = 0.0000
        # x(3) = 22.1408
        # y(3) = 0.0000
        # x(4) = 22.1185
        # y(4) = 2.5620
        self.assertAlmostEqual(-0.0190, round(area_shift.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(2.1755, round(area_shift.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_shift.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_shift.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(22.1408, round(area_shift.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_shift.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(22.1185, round(area_shift.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(2.5620, round(area_shift.get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0 (1-4) = nan
        # Alph0 (1-4) = 1.003
        # Phi0 (2-3) = 0.0000
        # Alph0 (2-3) = 0.0000
        self.assertAlmostEqual(1.0003, Line.calculate_relative_alpha_angle(area_shift.get_line(2)), places=4)

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 0.1323
        # y(1) = 2.0281
        # x(2) = 0.1487
        # y(2) = 0.1500
        # x(3) = 21.9895
        # y(3) = 0.1500
        # x(4) = 21.9698
        # y(4) = 2.4094

        area_bare = block.get_bare_isosceles_trapezium(p_ref=area_shift.get_line(0).p1,
                                                       alpha_ref=Line.calculate_relative_alpha_angle(
                                                           area_shift.get_line(0)))

        self.assertAlmostEqual(0.1323, round(area_bare.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(2.0281, round(area_bare.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(0.1487, round(area_bare.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.1500, round(area_bare.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(21.9895, round(area_bare.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(0.1500, round(area_bare.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(21.9698, round(area_bare.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(2.4094, round(area_bare.get_line(2).p1.y, 4), places=4)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    # BLOCK 1
    #     1     1     4           25            0            0        13500     16TIL9   2  20     0            0
    #    no  type   nco       radius          phi        alpha      current    condname  n1 n2  imag         turn
    #  INSUL 1
    #   1 16TINS2           0.15         0.15 'eurocircol                    '
    #  No Name            Radial       Azimut  Comment
    #
    #  CABLE 1
    #   1 16TOLG78B        21.84       1.8782       2.2595    38          100            3 'EUROCIRCOL 26 STR             '
    #  No Name            height      width_i      width_o    ns      transp.        degrd  Comment
    #
    # CONDUCTOR 1
    #   1 16TIL6E        1 16TOLG78B  STR11     NB3SN2    16TINS2   NONE      NB3SN                1.9 'eurocircol                    '
    #  No Name       Type CableGeom.  Strand    Filament  Insul     Trans     QuenchMat.           T_o  Comment
    @patch("matplotlib.pyplot.show")
    def test_get_first_isosceles_trapezium_traps_with_insulation_with_radius_no_phi_no_alpha(self, mock_show=None):
        block_rel_def = RelativeCosThetaBlockDefinition(radius=25, phi_r=0.0, alpha_r=0.0, nco=1, condname='16TIL6E',
                                                        no=1,
                                                        current=13500, type=1, n1=2, n2=20, imag=0, turn=0)
        block = RelativeCosThetaBlock(block_def=block_rel_def,
                                      cable_def=self.cadata.get_cable_definition('16TIL6E'),
                                      insul_def=self.cadata.get_insul_definition('16TIL6E'),
                                      strand_def=self.cadata.get_strand_definition('16TIL6E'),
                                      conductor_def=self.cadata.get_conductor_definition('16TIL6E'))

        p_ref = Point.of_polar(block.block_def.radius, block.get_phi())
        area_ins = block.get_insulated_isosceles_trapezium(p_ref=p_ref, alpha_ref=block.get_alpha())
        area_ins_shift = area_ins.copy()
        if block.is_area_intersecting_or_within_radius(area_ins):
            area_ins_shift = area_ins.translate(Point.of_polar(2 * block.cable_def.thickness_i, block.get_alpha()))
            shift = Area.find_trapezoid_shift_to_intercept_with_radius(block.block_def.radius, area_ins_shift, eps=1e-9)
            area_ins_shift = area_ins_shift.translate(area_ins.get_line(0).unit() * shift)

        fig, ax = plt.subplots(figsize=(15, 15))
        ax.grid(True)
        ax.set_xlim(24.9, 25.1)
        ax.set_ylim(0, 3)

        circle = plt.Circle((0, 0), block.block_def.radius, color='r', fill=False)
        ax.add_patch(circle)
        area_ins.plot(ax)
        area_ins_shift.plot(ax)
        plt.show()

        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 24.9820
        # y(1) = 2.1755
        # x(2) = 25.0010
        # y(2) = 0.0000
        # x(3) = 47.1418
        # y(3) = 0.0000
        # x(4) = 47.1194
        # y(4) = 2.5620
        self.assertAlmostEqual(24.9820, round(area_ins_shift.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(2.1755, round(area_ins_shift.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(25.0010, round(area_ins_shift.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_ins_shift.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(47.1418, round(area_ins_shift.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area_ins_shift.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(47.1194, round(area_ins_shift.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(2.5620, round(area_ins_shift.get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0(1 - 4) = 4.9891
        # Alph0(1 - 4) = 1.0003
        # Phi0(2 - 3) = 0.0000
        # Alph0(2 - 3) = 0.0000
        self.assertAlmostEqual(4.9891, round(
            Line.calculate_positioning_angle(area_ins_shift.get_line(2), block.block_def.radius), 4), places=4)
        self.assertAlmostEqual(1.0003, round(Line.calculate_relative_alpha_angle(area_ins_shift.get_line(2)), 4),
                               places=4)
        self.assertAlmostEqual(0.0000,
                               Line.calculate_positioning_angle(area_ins_shift.get_line(0), block.block_def.radius),
                               places=4)
        self.assertAlmostEqual(0.0000, Line.calculate_relative_alpha_angle(area_ins_shift.get_line(0)), places=4)

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 25.1333
        # y(1) = 2.0281
        # x(2) = 25.1496
        # y(2) = 0.1500
        # x(3) = 46.9905
        # y(3) = 0.1500
        # x(4) = 46.9708
        # y(4) = 2.4094

        area_bare = block.get_bare_isosceles_trapezium(p_ref=area_ins_shift.get_line(0).p1,
                                                       alpha_ref=Line.calculate_relative_alpha_angle(
                                                           area_ins_shift.get_line(0)))

        self.assertAlmostEqual(25.1333, round(area_bare.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(2.0281, round(area_bare.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(25.1496, round(area_bare.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.1500, round(area_bare.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(46.9905, round(area_bare.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(0.1500, round(area_bare.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(46.9708, round(area_bare.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(2.4094, round(area_bare.get_line(2).p1.y, 4), places=4)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    # BLOCK 1
    #     1     1     4           25            5            0        13500     16TIL9   2  20     0            0
    #    no  type   nco       radius          phi        alpha      current    condname  n1 n2  imag         turn
    #  INSUL 1
    #   1 16TINS2           0.15         0.15 'eurocircol                    '
    #  No Name            Radial       Azimut  Comment
    #
    #  CABLE 1
    #   1 16TOLG78B        21.84       1.8782       2.2595    38          100            3 'EUROCIRCOL 26 STR             '
    #  No Name            height      width_i      width_o    ns      transp.        degrd  Comment
    #
    # CONDUCTOR 1
    #   1 16TIL6E        1 16TOLG78B  STR11     NB3SN2    16TINS2   NONE      NB3SN                1.9 'eurocircol                    '
    #  No Name       Type CableGeom.  Strand    Filament  Insul     Trans     QuenchMat.           T_o  Comment
    @patch("matplotlib.pyplot.show")
    def test_get_first_isosceles_trapezium_trapz_with_insulation_with_radius_with_phi_no_alpha(self, mock_show=None):
        block_rel_def = RelativeCosThetaBlockDefinition(radius=25, phi_r=5.0, alpha_r=0.0, nco=1, condname='16TIL6E',
                                                        no=1,
                                                        current=13500, type=1, n1=2, n2=20, imag=0, turn=0)
        block = RelativeCosThetaBlock(block_def=block_rel_def,
                                      cable_def=self.cadata.get_cable_definition('16TIL6E'),
                                      insul_def=self.cadata.get_insul_definition('16TIL6E'),
                                      strand_def=self.cadata.get_strand_definition('16TIL6E'),
                                      conductor_def=self.cadata.get_conductor_definition('16TIL6E'))

        p_ref = Point.of_polar(block.block_def.radius, block.get_phi())
        area_ins = block.get_insulated_isosceles_trapezium(p_ref=p_ref, alpha_ref=block.get_alpha())
        area_ins_shift = area_ins.copy()
        if block.is_area_intersecting_or_within_radius(area_ins):
            area_ins_shift = area_ins.translate(Point.of_polar(2 * block.cable_def.thickness_i, block.get_alpha()))
            shift = Area.find_trapezoid_shift_to_intercept_with_radius(block.block_def.radius, area_ins_shift, eps=1e-6)
            area_ins_shift = area_ins_shift.translate(area_ins.get_line(0).unit() * shift)

        fig, ax = plt.subplots(figsize=(15, 15))
        ax.set_aspect('equal', 'box')
        ax.grid(True)
        ax.set_xlim(0, 75)
        ax.set_ylim(0, 75)

        circle = plt.Circle((0, 0), block.block_def.radius, color='r', fill=False)
        ax.add_patch(circle)
        area_ins.plot(ax)
        area_ins_shift.plot(ax)
        plt.show()

        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 24.8859
        # y(1) = 4.3544
        # x(2) = 24.9049
        # y(2) = 2.1789
        # x(3) = 47.0457
        # y(3) = 2.1789
        # x(4) = 47.0233
        # y(4) = 4.7409
        self.assertAlmostEqual(24.8859, round(area_ins_shift.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(4.3544, round(area_ins_shift.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(24.9049, round(area_ins_shift.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(2.1789, round(area_ins_shift.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(47.0457, round(area_ins_shift.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(2.1789, round(area_ins_shift.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(47.0233, round(area_ins_shift.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(4.7409, round(area_ins_shift.get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0(1 - 4) = 10.0199
        # Alph0(1 - 4) = 1.0003
        # Phi0(2 - 3) = 5.0000
        # Alph0(2 - 3) = 0.0000
        self.assertAlmostEqual(10.0199,
                               Line.calculate_positioning_angle(area_ins_shift.get_line(2), block.block_def.radius),
                               places=4)
        self.assertAlmostEqual(1.0003, Line.calculate_relative_alpha_angle(area_ins_shift.get_line(2)), places=4)
        self.assertAlmostEqual(5.0000,
                               Line.calculate_positioning_angle(area_ins_shift.get_line(0), block.block_def.radius),
                               places=4)
        self.assertAlmostEqual(0.0000, Line.calculate_relative_alpha_angle(area_ins_shift.get_line(0)), places=4)

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 25.0372
        # y(1) = 4.2070
        # x(2) = 25.0536
        # y(2) = 2.3289
        # x(3) = 46.8944
        # y(3) = 2.3289
        # x(4) = 46.8747
        # y(4) = 4.5883

        area_bare = block.get_bare_isosceles_trapezium(p_ref=area_ins_shift.get_line(0).p1,
                                                       alpha_ref=Line.calculate_relative_alpha_angle(
                                                           area_ins_shift.get_line(0)))

        self.assertAlmostEqual(25.0372, round(area_bare.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(4.2070, round(area_bare.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(25.0536, round(area_bare.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(2.3289, round(area_bare.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(46.8944, round(area_bare.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(2.3289, round(area_bare.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(46.8747, round(area_bare.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(4.5883, round(area_bare.get_line(2).p1.y, 4), places=4)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    # BLOCK 1
    #     1     1     4           25            5           10        13500     16TIL9   2  20     0            0
    #    no  type   nco       radius          phi        alpha      current    condname  n1 n2  imag         turn
    #  INSUL 1
    #   1 16TINS2           0.15         0.15 'eurocircol                    '
    #  No Name            Radial       Azimut  Comment
    #
    #  CABLE 1
    #   1 16TOLG78B        21.84       1.8782       2.2595    38          100            3 'EUROCIRCOL 26 STR             '
    #  No Name            height      width_i      width_o    ns      transp.        degrd  Comment
    #
    # CONDUCTOR 1
    #   1 16TIL6E        1 16TOLG78B  STR11     NB3SN2    16TINS2   NONE      NB3SN                1.9 'eurocircol                    '
    #  No Name       Type CableGeom.  Strand    Filament  Insul     Trans     QuenchMat.           T_o  Comment
    @patch("matplotlib.pyplot.show")
    def test_get_first_isosceles_trapezium_trapz_with_insulation_with_radius_with_phi_with_alpha(self, mock_show=None):
        block_rel_def = RelativeCosThetaBlockDefinition(radius=25, phi_r=5.0, alpha_r=10.0, nco=1, condname='16TIL6E',
                                                        no=1, current=13500, type=1, n1=2, n2=20, imag=0, turn=0)
        block = RelativeCosThetaBlock(block_def=block_rel_def,
                                      cable_def=self.cadata.get_cable_definition('16TIL6E'),
                                      insul_def=self.cadata.get_insul_definition('16TIL6E'),
                                      strand_def=self.cadata.get_strand_definition('16TIL6E'),
                                      conductor_def=self.cadata.get_conductor_definition('16TIL6E'))

        p_ref = Point.of_polar(block.block_def.radius, block.get_phi())
        area_ins = block.get_insulated_isosceles_trapezium(p_ref=p_ref, alpha_ref=block.get_alpha())
        area_ins_shift = area_ins.copy()
        if block.is_area_intersecting_or_within_radius(area_ins):
            area_ins_shift = area_ins.translate(Point.of_polar(2 * block.cable_def.thickness_i, block.get_alpha()))
            shift = Area.find_trapezoid_shift_to_intercept_with_radius(block.block_def.radius, area_ins_shift, eps=1e-6)
            area_ins_shift = area_ins_shift.translate(area_ins.get_line(0).unit() * shift)

        fig, ax = plt.subplots(figsize=(15, 15))
        ax.set_aspect('equal', 'box')
        ax.grid(True)
        ax.set_xlim(0, 75)
        ax.set_ylim(0, 75)

        circle = plt.Circle((0, 0), block.block_def.radius, color='r', fill=False)
        ax.add_patch(circle)
        area_ins.plot(ax)
        area_ins_shift.plot(ax)
        plt.show()

        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 24.6208
        # y(1) = 4.3379
        # x(2) = 25.0173
        # y(2) = 2.1987
        # x(3) = 46.8217
        # y(3) = 6.0434
        # x(4) = 46.3548
        # y(4) = 8.5627
        self.assertAlmostEqual(24.6208, round(area_ins_shift.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(4.3379, round(area_ins_shift.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(25.0173, round(area_ins_shift.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(2.1987, round(area_ins_shift.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(46.8217, round(area_ins_shift.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(6.0434, round(area_ins_shift.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(46.3548, round(area_ins_shift.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(8.5627, round(area_ins_shift.get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0(1 - 4) = 9.9922
        # Alph0(1 - 4) = 11.0003
        # Phi0(2 - 3) = 5.0000
        # Alph0(2 - 3) = 10.0000
        self.assertAlmostEqual(9.9922,
                               Line.calculate_positioning_angle(area_ins_shift.get_line(2), block.block_def.radius),
                               places=4)
        self.assertAlmostEqual(11.0003, Line.calculate_relative_alpha_angle(area_ins_shift.get_line(2)), places=4)
        self.assertAlmostEqual(5.0000,
                               Line.calculate_positioning_angle(area_ins_shift.get_line(0), block.block_def.radius),
                               places=4)
        self.assertAlmostEqual(10.0000, Line.calculate_relative_alpha_angle(area_ins_shift.get_line(0)), places=4)

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 24.7954
        # y(1) = 4.2190
        # x(2) = 25.1376
        # y(2) = 2.3723
        # x(3) = 46.6467
        # y(3) = 6.1649
        # x(4) = 46.2349
        # y(4) = 8.3865

        area_bare = block.get_bare_isosceles_trapezium(p_ref=area_ins_shift.get_line(0).p1,
                                                       alpha_ref=Line.calculate_relative_alpha_angle(
                                                           area_ins_shift.get_line(0)))

        self.assertAlmostEqual(24.7954, round(area_bare.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(4.2190, round(area_bare.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(25.1376, round(area_bare.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(2.3723, round(area_bare.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(46.6467, round(area_bare.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(6.1649, round(area_bare.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(46.2349, round(area_bare.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(8.3865, round(area_bare.get_line(2).p1.y, 4), places=4)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch.multiple(CosThetaBlock, __abstractmethods__=set())
    def test_get_alpha(self):
        # arrange
        cos_theta_block = CosThetaBlock(block_def=None,
                                        cable_def=None,
                                        insul_def=None,
                                        strand_def=None,
                                        conductor_def=None)

        # act
        # assert
        with self.assertRaises(NotImplementedError) as context:
            cos_theta_block.get_alpha()

        self.assertTrue('This method is not implemented for this class' in str(context.exception))

    @patch.multiple(CosThetaBlock, __abstractmethods__=set())
    def test_get_phi(self):
        # arrange
        cos_theta_block = CosThetaBlock(block_def=None,
                                        cable_def=None,
                                        insul_def=None,
                                        strand_def=None,
                                        conductor_def=None)

        # act
        # assert
        with self.assertRaises(NotImplementedError) as context:
            cos_theta_block.get_phi()

        self.assertTrue('This method is not implemented for this class' in str(context.exception))

    @patch.multiple(CosThetaBlock, __abstractmethods__=set())
    def test_to_roxie_df(self):
        # arrange
        cos_theta_block = CosThetaBlock(block_def=None,
                                        cable_def=None,
                                        insul_def=None,
                                        strand_def=None,
                                        conductor_def=None)

        # act
        # assert
        with self.assertRaises(NotImplementedError) as context:
            cos_theta_block.to_block_df()

        self.assertTrue('This method is not implemented for this class' in str(context.exception))

    def test_calculate_radiality_alpha_correction_index_error(self):
        # arrange
        block_rel_def = RelativeCosThetaBlockDefinition(radius=25, phi_r=5.0, alpha_r=10.0, nco=1, condname='16TIL6E',
                                                        no=1, current=13500, type=1, n1=2, n2=20, imag=0, turn=0)
        block = RelativeCosThetaBlock(block_def=block_rel_def,
                                      cable_def=self.cadata.get_cable_definition('16TIL6E'),
                                      insul_def=self.cadata.get_insul_definition('16TIL6E'),
                                      strand_def=self.cadata.get_strand_definition('16TIL6E'),
                                      conductor_def=self.cadata.get_conductor_definition('16TIL6E'))

        # act
        # assert
        with self.assertRaises(IndexError) as context:
            GeometryChange.compute_radial_alpha(block)

        exception_message = 'The list of areas is empty, please build block first with build_block() method.'
        self.assertTrue(exception_message in str(context.exception))

    def test_calculate_radiality_alpha_correction_single_turn(self):
        # arrange
        block_rel_def = RelativeCosThetaBlockDefinition(radius=25, phi_r=5.0, alpha_r=10.0, nco=1, condname='16TIL6E',
                                                        no=1, current=13500, type=1, n1=2, n2=20, imag=0, turn=0)
        block = RelativeCosThetaBlock(block_def=block_rel_def,
                                      cable_def=self.cadata.get_cable_definition('16TIL6E'),
                                      insul_def=self.cadata.get_insul_definition('16TIL6E'),
                                      strand_def=self.cadata.get_strand_definition('16TIL6E'),
                                      conductor_def=self.cadata.get_conductor_definition('16TIL6E'))

        # act
        block.build_block()

        # assert
        alpha_correction = GeometryChange.compute_radial_alpha(block)

        self.assertAlmostEqual(-2.99267610, alpha_correction, places=7)

    def test_calculate_radiality_alpha_correction_five_turns(self):
        # arrange
        block_rel_def = RelativeCosThetaBlockDefinition(radius=25, phi_r=5.0, alpha_r=10.0, nco=5, condname='16TIL6E',
                                                        no=1, current=13500, type=1, n1=2, n2=20, imag=0, turn=0)
        block = RelativeCosThetaBlock(block_def=block_rel_def,
                                      cable_def=self.cadata.get_cable_definition('16TIL6E'),
                                      insul_def=self.cadata.get_insul_definition('16TIL6E'),
                                      strand_def=self.cadata.get_strand_definition('16TIL6E'),
                                      conductor_def=self.cadata.get_conductor_definition('16TIL6E'))

        # act
        block.build_block()

        # assert
        alpha_correction = GeometryChange.compute_radial_alpha(block)

        self.assertAlmostEqual(4.93033090, alpha_correction, places=7)



import json
from typing import List, TypeVar

import pandas as pd

from magnumapi.cadata.CableDefinition import CableDefinition
from magnumapi.cadata.ConductorDefinition import ConductorDefinition
from magnumapi.cadata.FilamentDefinition import FilamentDefinition
from magnumapi.cadata.InsulationDefinition import InsulationDefinition
from magnumapi.cadata.QuenchDefinition import QuenchDefinition
from magnumapi.cadata.RemFitDefinition import RemFitDefinition
from magnumapi.cadata.StrandDefinition import StrandDefinition
from magnumapi.cadata.TransientDefinition import TransientDefinition
import magnumapi.tool_adapters.roxie.RoxieAPI as RoxieAPI
import magnumapi.commons.json_file as json_file
import magnumapi.commons.text_file as text_file
from magnumapi.commons.DirectoryManager import DirectoryManager

T = TypeVar('T')


class CableDatabase:
    """ Class providing an interface to read, write, and access cable database definitions.

    """
    keyword_to_class = {'INSUL': InsulationDefinition,
                        'REMFIT': RemFitDefinition,
                        'FILAMENT': FilamentDefinition,
                        'STRAND': StrandDefinition,
                        'TRANSIENT': TransientDefinition,
                        'QUENCH': QuenchDefinition,
                        'CABLE': CableDefinition,
                        'CONDUCTOR': ConductorDefinition}

    def __init__(self,
                 insul_defs: List[InsulationDefinition],
                 remfit_defs: List[RemFitDefinition],
                 filament_defs: List[FilamentDefinition],
                 strand_defs: List[StrandDefinition],
                 transient_defs: List[TransientDefinition],
                 quench_defs: List[QuenchDefinition],
                 cable_defs: List[CableDefinition],
                 conductor_defs: List[ConductorDefinition]) -> None:

        self.insul_defs = insul_defs
        self.remfit_defs = remfit_defs
        self.filament_defs = filament_defs
        self.strand_defs = strand_defs
        self.transient_defs = transient_defs
        self.quench_defs = quench_defs
        self.cable_defs = cable_defs
        self.conductor_defs = conductor_defs

    @classmethod
    def initialize_definitions(cls, cadata_file_path: str, keyword: str) -> List[T]:
        """ Method initializing a list of definitions of a given type from a given cadata file. Method reads a cadata
        file and returns a dictionary, which is converted into a list of dictionaries. The list of dictionaries is
        converted into a list of definitions.

        :param cadata_file_path: a path to a cadata file
        :param keyword: a cadata table name
        :return: a list of cadata definitions for a given table name
        """
        ClassDefinition = cls.keyword_to_class[keyword.upper()]
        try:
            df = RoxieAPI.read_bottom_header_table(cadata_file_path, keyword=keyword)
            df = df.drop(columns=['No'])
            df = df.rename(columns=ClassDefinition.get_roxie_to_magnum_dct())
            df_dicts = df.to_dict('records')
            defs = []
            for df_dict in df_dicts:
                defs.append(ClassDefinition(**df_dict))

            return defs
        except IndexError:
            return [ClassDefinition(name='', comment='')]

    def get_insul_definition(self, condname: str) -> InsulationDefinition:
        """ Method returning an insulation definition for a given conductor name

        :param condname: conductor name
        :return: insulation definition if match, otherwise a KeyError is thrown
        """
        insul_name = self.get_conductor_definition(condname).insulation

        return _find_matching_definition(self.insul_defs, insul_name, 'insulation')

    def get_remfit_definition(self, condname: str) -> RemFitDefinition:
        """ Method returning a remfit definition for a given conductor name

        :param condname: conductor name
        :return: remfit definition if match, otherwise a KeyError is thrown
        """
        remfit_name = self.get_filament_definition(condname).fit_perp

        return _find_matching_definition(self.remfit_defs, remfit_name, 'rem_fit')

    def get_filament_definition(self, condname: str) -> FilamentDefinition:
        """ Method returning a filament definition for a given conductor name

        :param condname: conductor name
        :return: filament definition if match, otherwise a KeyError is thrown
        """
        filament_name = self.get_conductor_definition(condname).filament

        return _find_matching_definition(self.filament_defs, filament_name, 'filament')

    def get_strand_definition(self, condname: str) -> StrandDefinition:
        """ Method returning a strand definition for a given conductor name

        :param condname: conductor name
        :return: strand definition if match, otherwise a KeyError is thrown
        """
        strand_name = self.get_conductor_definition(condname).strand

        return _find_matching_definition(self.strand_defs, strand_name, 'strand')

    def get_transient_definition(self, condname: str) -> TransientDefinition:
        """ Method returning a transient definition for a given conductor name. If there is no transient definition for
        a given condname, then an empty transient definition is returned.

        :param condname: conductor name
        :return: transient definition if match, otherwise a KeyError is thrown
        """
        trans_name = self.get_conductor_definition(condname).transient

        if trans_name == 'NONE':
            return TransientDefinition(name='', comment='')

        return _find_matching_definition(self.transient_defs, trans_name, 'transient')

    def get_quench_definition(self, condname: str) -> QuenchDefinition:
        """ Method returning a quench definition for a given conductor name. If there is no quench definition for
        a given condname, then an empty quench definition is returned.

        :param condname: conductor name
        :return: quench definition if match, otherwise a KeyError is thrown
        """
        quench_name = self.get_conductor_definition(condname).quench_mat

        if quench_name == 'NONE':
            return QuenchDefinition(name='', comment='')

        return _find_matching_definition(self.quench_defs, quench_name, 'quench')

    def get_cable_definition(self, condname: str) -> CableDefinition:
        """ Method returning an insulation definition for a given conductor name

        :param condname: conductor name
        :return: insulation definition if match, otherwise a KeyError is thrown
        """
        geometry_name = self.get_conductor_definition(condname).cable_geom

        return _find_matching_definition(self.cable_defs, geometry_name, 'cable')

    def get_conductor_definition(self, condname: str) -> ConductorDefinition:
        """ Method returning an insulation definition for a given conductor name

        :param condname: conductor name
        :return: insulation definition if match, otherwise a KeyError is thrown
        """
        return _find_matching_definition(self.conductor_defs, condname, 'conductor')

    @classmethod
    def read_json(cls, json_file_path: str) -> "CableDatabase":
        """ Method reading a json file and returning an initialized CableDatabase instance. Some definitions are
        optional. In this case, the

        :param json_file_path: a path
        :return: a CableDatabase instance with initialized lists of definitions
        """
        data = json_file.read(json_file_path)

        # Optional definitions
        if 'remfit' in data:
            remfit_defs = [RemFitDefinition(**remfit_def) for remfit_def in data['remfit']]
        else:
            remfit_defs = [RemFitDefinition(name='', comment='')]

        if 'transient' in data:
            transient_defs = [TransientDefinition(**transient_def) for transient_def in data['transient']]
        else:
            transient_defs = [TransientDefinition(name='', comment='')]

        if 'quench' in data:
            quench_defs = [QuenchDefinition(**quench_def) for quench_def in data['quench']]
        else:
            quench_defs = [QuenchDefinition(name='', comment='')]

        # Mandatory definitions
        insul_defs = [InsulationDefinition(**insulation_def) for insulation_def in data['insulation']]
        filament_defs = [FilamentDefinition(**filament_def) for filament_def in data['filament']]
        strand_defs = [StrandDefinition(**strand_def) for strand_def in data['strand']]
        cable_defs = [CableDefinition(**cable_def) for cable_def in data['cable']]
        conductor_defs = [ConductorDefinition(**conductor_def) for conductor_def in data['conductor']]

        return CableDatabase(insul_defs=insul_defs,
                             remfit_defs=remfit_defs,
                             filament_defs=filament_defs,
                             strand_defs=strand_defs,
                             transient_defs=transient_defs,
                             quench_defs=quench_defs,
                             cable_defs=cable_defs,
                             conductor_defs=conductor_defs)

    def write_json(self, json_output_path: str) -> None:
        """ Method writing a CableDatabase instance into a json file.

        :param json_output_path: a path to an output json file
        """
        json_cadata = {"insulation": self._get_insulation_definitions_as_list_of_dict(),
                       "remfit": self._get_remfit_definitions_as_list_of_dict(),
                       "filament": self._get_filament_definitions_as_list_of_dict(),
                       "strand": self._get_strand_definitions_as_list_of_dict(),
                       "transient": self._get_transient_definitions_as_list_of_dict(),
                       "quench": self._get_quench_definitions_as_list_of_dict(),
                       "cable": self._get_cable_definitions_as_list_of_dict(),
                       "conductor": self._get_conductor_definitions_as_list_of_dict()}

        with open(json_output_path, 'w', encoding='utf-8') as f:
            json.dump(json_cadata, f, ensure_ascii=False, indent=4)

    def _get_conductor_definitions_as_list_of_dict(self):
        """ Method returning a list of conductor definitions as a list of dictionaries sorted with ROXIE order.

        :return: a list of dictionaries with conductor definitions sorted with ROXIE order
        """
        return [ConductorDefinition.reorder_dct(conductor_def.__dict__) for conductor_def in self.conductor_defs]

    def _get_cable_definitions_as_list_of_dict(self):
        """ Method returning a list of cable definitions as a list of dictionaries sorted with ROXIE order.

        :return: a list of dictionaries with cable definitions sorted with ROXIE order
        """
        return [CableDefinition.reorder_dct(cable_def.__dict__) for cable_def in self.cable_defs]

    def _get_quench_definitions_as_list_of_dict(self):
        """ Method returning a list of quench definitions as a list of dictionaries sorted with ROXIE order.

        :return: a list of dictionaries with quench definitions sorted with ROXIE order
        """
        return [QuenchDefinition.reorder_dct(quench_def.__dict__) for quench_def in self.quench_defs]

    def _get_transient_definitions_as_list_of_dict(self):
        """ Method returning a list of transient definitions as a list of dictionaries sorted with ROXIE order.

        :return: a list of dictionaries with transient definitions sorted with ROXIE order
        """
        return [TransientDefinition.reorder_dct(transient_def.__dict__) for transient_def in self.transient_defs]

    def _get_strand_definitions_as_list_of_dict(self):
        """ Method returning a list of strand definitions as a list of dictionaries sorted with ROXIE order.

        :return: a list of dictionaries with strand definitions sorted with ROXIE order
        """
        return [StrandDefinition.reorder_dct(strand_def.__dict__) for strand_def in self.strand_defs]

    def _get_filament_definitions_as_list_of_dict(self):
        """ Method returning a list of filament definitions as a list of dictionaries sorted with ROXIE order.

        :return: a list of dictionaries with filament definitions sorted with ROXIE order
        """
        return [FilamentDefinition.reorder_dct(filament_def.__dict__) for filament_def in self.filament_defs]

    def _get_remfit_definitions_as_list_of_dict(self):
        """ Method returning a list of remfit definitions as a list of dictionaries sorted with ROXIE order.

        :return: a list of dictionaries with remfit definitions sorted with ROXIE order
        """
        return [RemFitDefinition.reorder_dct(remfit_def.__dict__) for remfit_def in self.remfit_defs]

    def _get_insulation_definitions_as_list_of_dict(self):
        """ Method returning a list of insulation definitions as a list of dictionaries sorted with ROXIE order.

        :return: a list of dictionaries with insulation definitions sorted with ROXIE order
        """
        return [InsulationDefinition.reorder_dct(insul_def.__dict__) for insul_def in self.insul_defs]

    @classmethod
    def read_cadata(cls, cadata_file_path: str) -> "CableDatabase":
        """ Method reading a cadata file and returns an initialized CableDatabase instance with a list of definitions.

        :param cadata_file_path: a path to a cadata file
        :return: a CableDatabase instance with initialized lists of definitions
        """
        DirectoryManager.check_if_file_exists(cadata_file_path)

        return CableDatabase(insul_defs=cls.initialize_definitions(cadata_file_path, keyword='INSUL'),
                             remfit_defs=cls.initialize_definitions(cadata_file_path, keyword='REMFIT'),
                             filament_defs=cls.initialize_definitions(cadata_file_path, keyword='FILAMENT'),
                             strand_defs=cls.initialize_definitions(cadata_file_path, keyword='STRAND'),
                             transient_defs=cls.initialize_definitions(cadata_file_path, keyword='TRANSIENT'),
                             quench_defs=cls.initialize_definitions(cadata_file_path, keyword='QUENCH'),
                             cable_defs=cls.initialize_definitions(cadata_file_path, keyword='CABLE'),
                             conductor_defs=cls.initialize_definitions(cadata_file_path, keyword='CONDUCTOR'))

    def write_cadata(self, cadata_output_path: str) -> None:
        """ Method writing a CableDatabase instance into a cadata file.

        :param cadata_output_path: a path to an output cadata file
        """
        output = ['VERSION 11',
                  self._convert_definition_df_to_bottom_header_str(self.get_insul_df(), 'INSUL'),
                  self._convert_definition_df_to_bottom_header_str(self.get_remfit_df(), 'REMFIT'),
                  self._convert_definition_df_to_bottom_header_str(self.get_filament_df(), 'FILAMENT'),
                  self._convert_definition_df_to_bottom_header_str(self.get_strand_df(), 'STRAND'),
                  self._convert_definition_df_to_bottom_header_str(self.get_transient_df(), 'TRANSIENT'),
                  self._convert_definition_df_to_bottom_header_str(self.get_quench_df(), 'QUENCH'),
                  self._convert_definition_df_to_bottom_header_str(self.get_cable_df(), 'CABLE'),
                  self._convert_definition_df_to_bottom_header_str(self.get_conductor_df(), 'CONDUCTOR')
                  ]

        # Write to a text file
        text_file.writelines(cadata_output_path, output, endline='\n\n')

    @classmethod
    def _convert_definition_df_to_bottom_header_str(cls, df: pd.DataFrame, keyword: str) -> str:
        """ Method converting a definition dataframe to a string representation as a bottom header table for ROXIE.

        :param df: input dataframe with definitions
        :param keyword: name of a table
        :return: a string formatted as a bottom header table of ROXIE
        """
        # Get the definition class
        ClassDefinition = cls.keyword_to_class[keyword.upper()]
        # Convert to a dataframe
        df = df.rename(columns=ClassDefinition.get_magnum_to_roxie_dct())
        # Take only those columns that are needed for ROXIE
        df = df[ClassDefinition.get_roxie_to_magnum_dct().keys()]
        # Add apostrophes around comment column
        df['Comment'] = '\'' + df['Comment'] + '\''
        # Add No column (1-based)
        columns = df.columns
        df['No'] = df.index + 1
        df = df[['No'] + list(columns)]
        df = df.astype({'No': 'int32'})
        # Convert a dataframe to a bottom header table
        return RoxieAPI.convert_bottom_header_table_to_str(df, keyword=keyword)

    def get_insul_df(self) -> pd.DataFrame:
        """ Method returning a dataframe table with insulation definitions.

        :return: a dataframe table with insulation definitions.
        """
        insulation_definitions = self._get_insulation_definitions_as_list_of_dict()
        return pd.DataFrame(insulation_definitions)

    def get_remfit_df(self) -> pd.DataFrame:
        """ Method returning a dataframe table with remfit definitions.

        :return: a dataframe table with remfit definitions.
        """
        remfit_definitions = self._get_remfit_definitions_as_list_of_dict()
        return pd.DataFrame(remfit_definitions)

    def get_filament_df(self) -> pd.DataFrame:
        """ Method returning a dataframe table with filament definitions.

        :return: a dataframe table with filament definitions.
        """
        filament_definitions = self._get_filament_definitions_as_list_of_dict()
        return pd.DataFrame(filament_definitions)

    def get_strand_df(self) -> pd.DataFrame:
        """ Method returning a dataframe table with strand definitions.

        :return: a dataframe table with strand definitions.
        """
        strand_definitions = self._get_strand_definitions_as_list_of_dict()
        return pd.DataFrame(strand_definitions)

    def get_transient_df(self) -> pd.DataFrame:
        """ Method returning a dataframe table with transient definitions.

        :return: a dataframe table with transient definitions.
        """
        transient_definitions = self._get_transient_definitions_as_list_of_dict()
        return pd.DataFrame(transient_definitions)

    def get_quench_df(self) -> pd.DataFrame:
        """ Method returning a dataframe table with quench definitions.

        :return: a dataframe table with quench definitions.
        """
        quench_definitions = self._get_quench_definitions_as_list_of_dict()
        return pd.DataFrame(quench_definitions)

    def get_cable_df(self) -> pd.DataFrame:
        """ Method returning a dataframe table with cable definitions.

        :return: a dataframe table with cable definitions.
        """
        cable_definitions = self._get_cable_definitions_as_list_of_dict()
        return pd.DataFrame(cable_definitions)

    def get_conductor_df(self) -> pd.DataFrame:
        """ Method returning a dataframe table with conductor definitions.

        :return: a dataframe table with conductor definitions.
        """
        conductor_definitions = self._get_conductor_definitions_as_list_of_dict()
        return pd.DataFrame(conductor_definitions)


def _find_matching_definition(defs: List[T], name_def: str, desc_def: str) -> T:
    """ Function finding a definition with a matching name in a list of input definitions. If there is a match, then
    the definition is returned, otherwise a KeyError is thrown.

    :param defs: list of input definitions over which a search is performed.
    :param name_def: name of a definition to find.
    :param desc_def: name of a key storing the definition name.
    :return: If there is a match, then the definition is returned, otherwise a KeyError is thrown.
    """
    matches = list(filter(lambda x: x.name == name_def, defs))

    if matches:
        return matches[0]
    else:
        raise KeyError('%s name %s not present in %s definitions.' % (desc_def.capitalize(), name_def, desc_def))



import unittest
from unittest.mock import patch

import pandas as pd

from magnumapi.geometry.GeometryFactory import GeometryFactory
from magnumapi.geometry.primitives.Line import Line
from magnumapi.geometry.blocks.RectangularBlock import RectangularBlock
from magnumapi.cadata.CableDatabase import CableDatabase
from tests.resource_files import read_csv_as_pd, create_resources_file_path


class TestRectangularGeometry(unittest.TestCase):

    @patch("matplotlib.pyplot.show")
    def test_build_geometry_rectangle_with_dict(self, mock_show=None):
        # arrange
        block_def = [
            {'no': 1, 'radius': 40.0, 'phi': 0, 'alpha': 0.0, 'nco': 3, 'type': 2, 'current': 13500,
             'condname': 'SMC11T100', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0}]
        layer_def = [{"no": 1,"symm": 1,"typexy": 1,"blocks": [1]}]

        cadata_path = create_resources_file_path('resources/geometry/roxie/SMC/roxie-SMC.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_dict(block_def, layer_def, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()

        # assert
        geometry_ref = read_csv_as_pd('resources/geometry/roxie/SMC/reference/three_cable_coil.csv')
        pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        # assert - compare to ROXIE
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 40.0000
        # y(1) = 1.5050
        # x(2) = 40.0000
        # y(2) = 0.0000
        # x(3) = 55.0470
        # y(3) = 0.0000
        # x(4) = 55.0470
        # y(4) = 1.5050
        area = geometry.blocks[0].areas[0]
        self.assertAlmostEqual(40.0000, round(area.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(1.5050, round(area.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(40.0000, round(area.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(55.0470, round(area.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(55.0470, round(area.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(1.5050, round(area.get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0 (1-4) = 2.1547
        # Alph0 (1-4) = 0.0000
        # Phi0 (2-3) = 0.0000
        # Alph0 (2-3) = 0.0000
        self.assertAlmostEqual(2.1547, area.get_line(2).p2.get_phi(), places=4)
        self.assertAlmostEqual(0.0000, Line.calculate_relative_alpha_angle(area.get_line(2)), places=4)
        self.assertAlmostEqual(0.0000, area.get_line(0).p1.get_phi(), places=4)
        self.assertAlmostEqual(0.0000, Line.calculate_relative_alpha_angle(area.get_line(0)), places=4)

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 40.1000
        # y(1) = 1.4050
        # x(2) = 40.1000
        # y(2) = 0.1000
        # x(3) = 54.9470
        # y(3) = 0.1000
        # x(4) = 54.9470
        # y(4) = 1.4050

        area_bare = geometry.blocks[0].get_bare_area(area)

        self.assertAlmostEqual(40.1000, round(area_bare.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(1.4050, round(area_bare.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(40.1000, round(area_bare.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.1000, round(area_bare.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(54.9470, round(area_bare.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(0.1000, round(area_bare.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(54.9470, round(area_bare.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(1.4050, round(area_bare.get_line(2).p1.y, 4), places=4)

        # assert - compare to ROXIE
        # Conductor number: 2
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 40.0000
        # y(1) = 3.0100
        # x(2) = 40.0000
        # y(2) = 1.5050
        # x(3) = 55.0470
        # y(3) = 1.5050
        # x(4) = 55.0470
        # y(4) = 3.0100
        area = geometry.blocks[0].areas[1]
        self.assertAlmostEqual(40.0000, round(area.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(3.0100, round(area.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(40.0000, round(area.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(1.5050, round(area.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(55.0470, round(area.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(1.5050, round(area.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(55.0470, round(area.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(3.0100, round(area.get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0 (1-4) = 4.3034
        # Alph0 (1-4) = 0.0000
        # Phi0 (2-3) = 2.1547
        # Alph0 (2-3) = 0.0000
        self.assertAlmostEqual(4.3034, area.get_line(2).p2.get_phi(), places=4)
        self.assertAlmostEqual(0.0000, Line.calculate_relative_alpha_angle(area.get_line(2)), places=4)
        self.assertAlmostEqual(2.1547, area.get_line(0).p1.get_phi(), places=4)
        self.assertAlmostEqual(0.0000, Line.calculate_relative_alpha_angle(area.get_line(0)), places=4)

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 40.1000
        # y(1) = 2.9100
        # x(2) = 40.1000
        # y(2) = 1.6050
        # x(3) = 54.9470
        # y(3) = 1.6050
        # x(4) = 54.9470
        # y(4) = 2.9100

        area_bare = geometry.blocks[0].get_bare_area(area)

        self.assertAlmostEqual(40.1000, round(area_bare.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(2.9100, round(area_bare.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(40.1000, round(area_bare.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(1.6050, round(area_bare.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(54.9470, round(area_bare.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(1.6050, round(area_bare.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(54.9470, round(area_bare.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(2.9100, round(area_bare.get_line(2).p1.y, 4), places=4)

        # assert - compare to ROXIE
        # Conductor number: 3
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 40.0000
        # y(1) = 4.5150
        # x(2) = 40.0000
        # y(2) = 3.0100
        # x(3) = 55.0470
        # y(3) = 3.0100
        # x(4) = 55.0470
        # y(4) = 4.5150
        area = geometry.blocks[0].areas[2]
        self.assertAlmostEqual(40.0000, round(area.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(4.5150, round(area.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(40.0000, round(area.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(3.0100, round(area.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(55.0470, round(area.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(3.0100, round(area.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(55.0470, round(area.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(4.5150, round(area.get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0 (1-4) = 6.4400
        # Alph0 (1-4) = 0.0000
        # Phi0 (2-3) = 4.3034
        # Alph0 (2-3) = 0.0000
        self.assertAlmostEqual(6.4400, area.get_line(2).p2.get_phi(), places=4)
        self.assertAlmostEqual(0.0000, Line.calculate_relative_alpha_angle(area.get_line(2)), places=4)
        self.assertAlmostEqual(4.3034, area.get_line(0).p1.get_phi(), places=4)
        self.assertAlmostEqual(0.0000, Line.calculate_relative_alpha_angle(area.get_line(0)), places=4)

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 40.1000
        # y(1) = 4.4150
        # x(2) = 40.1000
        # y(2) = 3.1100
        # x(3) = 54.9470
        # y(3) = 3.1100
        # x(4) = 54.9470
        # y(4) = 4.4150

        area_bare = geometry.blocks[0].get_bare_area(area)

        self.assertAlmostEqual(40.1000, round(area_bare.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(4.4150, round(area_bare.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(40.1000, round(area_bare.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(3.1100, round(area_bare.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(54.9470, round(area_bare.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(3.1100, round(area_bare.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(54.9470, round(area_bare.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(4.4150, round(area_bare.get_line(2).p1.y, 4), places=4)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_build_geometry_rectangle_with_json(self, mock_show=None):
        # arrange
        json_path = create_resources_file_path('resources/geometry/roxie/SMC/three_cable_coil.json')
        cadata_path = create_resources_file_path('resources/geometry/roxie/SMC/roxie-SMC.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_json(json_path, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()

        # assert
        geometry_ref = read_csv_as_pd('resources/geometry/roxie/SMC/reference/three_cable_coil.csv')
        pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_build_geometry_rectangle_with_data(self, mock_show=None):
        # arrange
        data_path = create_resources_file_path('resources/geometry/roxie/SMC/three_cable_coil.data')
        cadata_path = create_resources_file_path('resources/geometry/roxie/SMC/roxie-SMC.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_data(data_path, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()

        # assert
        geometry_ref = read_csv_as_pd('resources/geometry/roxie/SMC/reference/three_cable_coil.csv')
        pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_build_geometry_rectangle_with_csv(self, mock_show=None):
        # arrange
        block_dl_path = create_resources_file_path('resources/geometry/roxie/SMC/three_cable_coil.csv')
        layer_df_path = create_resources_file_path('resources/geometry/roxie/SMC/layer_data.csv')
        cadata_path = create_resources_file_path('resources/geometry/roxie/SMC/roxie-SMC.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_csv(block_dl_path, layer_df_path, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()

        # assert
        geometry_ref = read_csv_as_pd('resources/geometry/roxie/SMC/reference/three_cable_coil.csv')
        pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_smc_build_geometry_rectangle_with_data(self, mock_show=None):
        # arrange
        data_path = create_resources_file_path('resources/geometry/roxie/SMC/SMC11T_v3_35turns_ins100_new.data')
        cadata_path = create_resources_file_path('resources/geometry/roxie/SMC/roxie-SMC.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_data(data_path, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()

        # assert
        geometry_ref = read_csv_as_pd('resources/geometry/roxie/SMC/reference/SMC11T_v3_35turns_ins100_new.csv')
        pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("plotly.graph_objects.Figure.show")
    def test_smc_plotly_geometry_rectangle_with_data(self, mock_show=None):
        # arrange
        data_path = create_resources_file_path('resources/geometry/roxie/SMC/SMC11T_v3_35turns_ins100_new.data')
        cadata_path = create_resources_file_path('resources/geometry/roxie/SMC/roxie-SMC.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_data(data_path, cadata)
        geometry.build_blocks()
        geometry.plotly_blocks()

        # assert
        geometry_ref = read_csv_as_pd('resources/geometry/roxie/SMC/reference/SMC11T_v3_35turns_ins100_new.csv')
        pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    def test_to_roxie_df_with_csv(self):
        # arrange
        block_df_path = create_resources_file_path('resources/geometry/roxie/SMC/three_cable_coil.csv')
        layer_df_path = create_resources_file_path('resources/geometry/roxie/SMC/layer_data.csv')
        cadata_path = create_resources_file_path('resources/geometry/roxie/SMC/roxie-SMC.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_csv(block_df_path, layer_df_path, cadata)

        # assert
        roxie_ref = read_csv_as_pd('resources/geometry/roxie/SMC/three_cable_coil.csv')
        roxie_ref = roxie_ref.rename(columns={val: key for key, val in RectangularBlock.roxie_to_magnum_dct.items()})
        pd.testing.assert_frame_equal(roxie_ref.reset_index(drop=True), geometry.to_block_df().reset_index(drop=True))



from dataclasses import dataclass


@dataclass
class OptimizationNotebookConfig:
    """Class for optimization notebook config used for parametrizing notebooks used for optimization.

    Attributes:
       notebook_folder (str): The name of the root folder with the notebook
       notebook_name (str): The name of the notebook
       input_parameters (dict): The dictionary of input notebook parameters. The key is the variable name inside the
       notebook, while the value is the name of the exposed variable from another notebook
       output_parameters (list): The list of output notebook parameters
       input_artefacts (int): A dictionary of input artefacts for a notebook. The key is the path of an artefact for
       the notebook that takes this artefact as an input. The value is the output path from a notebook generating the
       artefact (w.r.t. the root directory).
       output_artefacts (list): The list of output artefacts produced by a given notebook.
    """
    notebook_folder: str
    notebook_name: str
    input_parameters: dict
    output_parameters: list
    input_artefacts: dict
    output_artefacts: list

    def __str__(self):
        return "notebook_folder: %s\n" \
               "notebook_name: %s\n" \
               "input_parameters: %s\n" \
               "output_parameters: %s\n" \
               "input_artefacts: %s\n" \
               "output_artefacts: %s" % (self.notebook_folder,
                                         self.notebook_name,
                                         self.input_parameters,
                                         self.output_parameters,
                                         self.input_artefacts,
                                         self.output_artefacts)



from unittest import TestCase

from magnumapi.optimization.OptimizationConfig import OptimizationConfig
from tests.resource_files import create_resources_file_path


class TestOptimizationConfig(TestCase):
    def setUp(self) -> None:
        json_path = create_resources_file_path('resources/optimization/config.json')
        self.config = OptimizationConfig.initialize_config(json_path)

    def test_initialize_config(self):
        # arrange
        # act
        # assert
        self.assertEqual('/home/mmaciejewski/gitlab/magnum-nb/', self.config.input_folder)
        self.assertEqual('/home/mmaciejewski/gitlab/magnum-nb/output/', self.config.output_folder)
        self.assertEqual('optimization/GeneticOptimization.csv', self.config.logger_rel_path)
        self.assertEqual('programmable', self.config.model_creation_type)

        self.assertEqual(100, self.config.n_gen)
        self.assertEqual(20, self.config.n_pop)
        self.assertAlmostEqual(0.9, self.config.r_cross, places=1)
        self.assertAlmostEqual(0.01, self.config.r_mut, places=2)

        self.assertEqual('b3', self.config.objectives[0].objective)
        self.assertAlmostEqual(0.1, self.config.objectives[0].weight, places=1)
        self.assertAlmostEqual(0.0, self.config.objectives[0].constraint, places=1)

        self.assertEqual('b5', self.config.objectives[1].objective)
        self.assertAlmostEqual(0.1, self.config.objectives[1].weight, places=1)
        self.assertAlmostEqual(0.0, self.config.objectives[1].constraint, places=1)

        self.assertEqual('margmi', self.config.objectives[2].objective)
        self.assertAlmostEqual(1, self.config.objectives[2].weight, places=0)
        self.assertAlmostEqual(0.85, self.config.objectives[2].constraint, places=2)

        self.assertEqual('seqv', self.config.objectives[3].objective)
        self.assertAlmostEqual(0.001, self.config.objectives[3].weight, places=3)
        self.assertAlmostEqual(0.0, self.config.objectives[3].constraint, places=1)

        self.assertEqual('geometry', self.config.notebooks[0].notebook_folder)
        self.assertEqual('Geometry.ipynb', self.config.notebooks[0].notebook_name)
        self.assertEqual({}, self.config.notebooks[0].input_parameters)
        self.assertEqual([], self.config.notebooks[0].output_parameters)
        self.assertEqual({}, self.config.notebooks[0].input_artefacts)
        self.assertEqual([], self.config.notebooks[0].output_artefacts)

        self.assertEqual('magnetic', self.config.notebooks[1].notebook_folder)
        self.assertEqual('ROXIE.ipynb', self.config.notebooks[1].notebook_name)
        self.assertEqual({}, self.config.notebooks[1].input_parameters)
        self.assertEqual(['b3', 'b5', 'bigb', 'margmi'], self.config.notebooks[1].output_parameters)
        self.assertEqual({}, self.config.notebooks[1].input_artefacts)
        self.assertEqual(['magnetic/input/roxie_scaled.force2d'],
                         self.config.notebooks[1].output_artefacts)

        self.assertEqual('mechanical', self.config.notebooks[2].notebook_folder)
        self.assertEqual('ANSYS.ipynb', self.config.notebooks[2].notebook_name)
        self.assertEqual({}, self.config.notebooks[2].input_parameters)
        self.assertEqual(['seqv', 'sxmn', 'syin', 'symn', 'syou'], self.config.notebooks[2].output_parameters)
        self.assertEqual({'input/roxie_scaled.force2d': 'magnetic/input/roxie_scaled.force2d'},
                         self.config.notebooks[2].input_artefacts)
        self.assertEqual([], self.config.notebooks[2].output_artefacts)

        self.assertEqual('thermal', self.config.notebooks[3].notebook_folder)
        self.assertEqual('MIITs.ipynb', self.config.notebooks[3].notebook_name)
        self.assertEqual({'peak_field': 'bigb'}, self.config.notebooks[3].input_parameters)
        self.assertEqual(['T_hotspot'], self.config.notebooks[3].output_parameters)
        self.assertEqual({}, self.config.notebooks[3].input_artefacts)
        self.assertEqual([], self.config.notebooks[3].output_artefacts)

    def test_get_weight(self):
        # arrange
        objective = 'b3'

        # act
        weight = self.config.get_weight(objective)

        # assert
        self.assertAlmostEqual(0.1, weight)

    def test_get_weight_error(self):
        # arrange
        objective = 'b7'

        # act
        with self.assertRaises(KeyError) as context:
            self.config.get_weight(objective)

        # assert
        self.assertTrue('Objective name b7 not present in objective configs.' in str(context.exception))

    def test_get_constraint(self):
        # arrange
        objective = 'b3'

        # act
        constraint = self.config.get_constraint(objective)

        # assert
        self.assertAlmostEqual(0, constraint)

    def test_get_constraint_error(self):
        # arrange
        objective = 'b7'

        # act
        with self.assertRaises(KeyError) as context:
            self.config.get_constraint(objective)

        # assert
        self.assertTrue('Objective name b7 not present in objective configs.' in str(context.exception))



import unittest
from unittest.mock import patch

from magnumapi.geometry.GeometryFactory import GeometryFactory
from magnumapi.cadata.CableDatabase import CableDatabase
from magnumapi.tool_adapters.ansys.AnsysInputBuilder import AnsysInputBuilder
import magnumapi.commons.text_file as text_file
from tests.resource_files import create_resources_file_path


class TestAnsysInputBuilder(unittest.TestCase):
    @patch("matplotlib.pyplot.show")
    def test_input_file_generation(self, mock_show=None):

        json_path = create_resources_file_path('resources/geometry/roxie/16T/16T_rel.json')
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        geometry = GeometryFactory.init_with_json(json_path, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()

        homo_geometry = geometry.homogenize()

        # Find number of layers
        n_layers = homo_geometry.get_number_of_layers()

        self.assertEqual(4, n_layers)
        # For each layer
        # # Number of blocks per layer
        blocks_per_layer = homo_geometry.get_number_of_blocks_per_layer()
        self.assertListEqual([4, 3, 3, 2], blocks_per_layer)

        output_text = AnsysInputBuilder.generate_ansys_input_text(homo_geometry)
        output_text_ref_path = create_resources_file_path('resources/tool_adapters/ansys/Model.inp')
        output_text_ref = text_file.readlines(output_text_ref_path)

        self.assertListEqual(output_text_ref, output_text)
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_input_file_generation_sm_ct(self, mock_show=None):

        json_path = create_resources_file_path('resources/geometry/roxie/16T/16T_rel_slotted.json')
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        geometry = GeometryFactory.init_slotted_with_json(json_path, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()

        homo_geometry = geometry.homogenize()

        output_text = AnsysInputBuilder.generate_ansys_input_text_sm_ct(homo_geometry)
        output_text_ref_path = create_resources_file_path('resources/tool_adapters/ansys/Model_slotted.inp')
        output_text_ref = text_file.readlines(output_text_ref_path)
        output_text_ref = [el.replace('\n', '') for el in output_text_ref]
        self.assertListEqual(output_text_ref, output_text)
        if mock_show is not None:
            mock_show.assert_called()

    def test_update_ansys_input_template_15T_mech(self):
        # arrange
        template_path = create_resources_file_path('resources/tool_adapters/ansys/15T_mech.template')
        output_dir = create_resources_file_path('resources/tool_adapters/ansys/')
        index = 0

        # act
        AnsysInputBuilder.update_input_template(template_path, index, output_dir, '15T_mech_%d.inp')

        # assert
        inp_path = create_resources_file_path('resources/tool_adapters/ansys/15T_mech_0.inp')
        inp_ref_path = create_resources_file_path('resources/tool_adapters/ansys/15T_mech_0_ref.inp')

        # # read reference
        inp_ref = text_file.readlines(inp_ref_path)

        # # read created file
        inp = text_file.readlines(inp_path)

        # # compare two strings
        assert inp_ref == inp

    def test_update_ansys_input_template_15T_mech_post_roxie(self):
        # arrange
        template_path = create_resources_file_path('resources/tool_adapters/ansys/15T_mech_post_roxie.template')
        output_dir = create_resources_file_path('resources/tool_adapters/ansys/')
        index = 0

        # act
        AnsysInputBuilder.update_input_template(template_path, index, output_dir, '15T_mech_post_roxie_%d.inp')

        # assert
        inp_path = create_resources_file_path('resources/tool_adapters/ansys/15T_mech_post_roxie_0.inp')
        inp_ref_path = create_resources_file_path('resources/tool_adapters/ansys/15T_mech_post_roxie_0_ref.inp')

        # # read reference
        inp_ref = text_file.readlines(inp_ref_path)

        # # read created file
        inp = text_file.readlines(inp_path)

        # # compare two strings
        assert inp_ref == inp

    def test_update_ansys_input_template_15T_mech_solu(self):
        # arrange
        template_path = create_resources_file_path('resources/tool_adapters/ansys/15T_mech_solu.template')
        output_dir = create_resources_file_path('resources/tool_adapters/ansys/')
        index = 0

        # act
        AnsysInputBuilder.update_input_template(template_path, index, output_dir, '15T_mech_solu_%d.inp')

        # assert
        inp_path = create_resources_file_path('resources/tool_adapters/ansys/15T_mech_solu_0.inp')
        inp_ref_path = create_resources_file_path('resources/tool_adapters/ansys/15T_mech_solu_0_ref.inp')

        # # read reference
        inp_ref = text_file.readlines(inp_ref_path)

        # # read created file
        inp = text_file.readlines(inp_path)

        # # compare two strings
        assert inp_ref == inp



import math
from unittest import TestCase

import dataclasses

from magnumapi.geometry.primitives.Line import Line
from magnumapi.geometry.primitives.Point import Point


class TestLine(TestCase):
    def test_calc_arc_length_between_two_lines(self):
        # arrange
        radius = 10
        center = Point.of_cartesian(0, 0)
        p_end_first = Point.of_polar(20, 30)
        p_end_second = Point.of_polar(20, 60)

        l_first = Line.of_end_points(center, p_end_first)
        l_second = Line.of_end_points(center, p_end_second)

        # act
        l_arc = Line.calc_arc_length_between_two_lines(radius, l_first, l_second)

        # assert
        self.assertAlmostEqual(radius * math.radians(30), l_arc, places=6)

    def test_calculate_point_orientation_wrt_line_positive_on_left(self):
        # arrange
        p1 = Point.of_cartesian(0, 0)
        p2 = Point.of_polar(20, 30)

        point = Point.of_polar(30, 45)
        line = Line.of_end_points(p1, p2)

        # act
        orientation = Line.calculate_point_orientation_wrt_line(line, point)

        # assert
        self.assertEqual(1, orientation)

    def test_calculate_point_orientation_wrt_line_negative_on_right(self):
        # arrange
        p1 = Point.of_cartesian(0, 0)
        p2 = Point.of_polar(20, 30)

        point = Point.of_polar(30, 15)
        line = Line.of_end_points(p1, p2)

        # act
        orientation = Line.calculate_point_orientation_wrt_line(line, point)

        # assert
        self.assertEqual(-1, orientation)

    def test_calculate_point_orientation_wrt_line_zero_on_line(self):
        # arrange
        p1 = Point.of_cartesian(0, 0)
        p2 = Point.of_polar(20, 30)

        point = Point.of_polar(10, 30)
        line = Line.of_end_points(p1, p2)

        # act
        orientation = Line.calculate_point_orientation_wrt_line(line, point)

        # assert
        self.assertEqual(0, orientation)

    def test_rotate(self):
        # arrange
        p1 = Point.of_polar(10, 10)
        p2 = Point.of_polar(20, 30)

        line = Line.of_end_points(p1, p2)

        alpha = 30

        # act
        line_rotated = line.rotate(alpha)

        # assert
        self.assertAlmostEqual(Point.of_polar(10, 40).x, line_rotated.p1.x, places=6)
        self.assertAlmostEqual(Point.of_polar(10, 40).y, line_rotated.p1.y, places=6)
        self.assertAlmostEqual(Point.of_polar(20, 60).x, line_rotated.p2.x, places=6)
        self.assertAlmostEqual(Point.of_polar(20, 60).y, line_rotated.p2.y, places=6)

    def test_str(self):
        # arrange
        p1 = Point.of_polar(10, 10)
        p2 = Point.of_polar(20, 30)

        line = Line.of_end_points(p1, p2)

        # act
        str_act = str(line)

        # assert
        self.assertEqual("(9.848078, 1.736482) -> (17.320508, 10.000000)", str_act)

    def test_immutability_p1(self):
        # arrange
        p1 = Point.of_polar(10, 10)
        p2 = Point.of_polar(20, 30)


        # act
        line = Line.of_end_points(p1, p2)

        # act
        with self.assertRaises(dataclasses.FrozenInstanceError):
            exec('line.p1 = p2')

    def test_immutability_p2(self):
        # arrange
        p1 = Point.of_polar(10, 10)
        p2 = Point.of_polar(20, 30)


        # act
        line = Line.of_end_points(p1, p2)

        # act
        with self.assertRaises(dataclasses.FrozenInstanceError):
            exec('line.p2 = p1')



import math

from magnumapi.cadata.StrandDefinition import StrandDefinition

strand_def = StrandDefinition(name='STR01',
                              d_strand=1.065,
                              f_cu_nocu=1.6,
                              rrr=70,
                              temp_ref=1.9,
                              b_ref=10,
                              j_c_at_b_ref_t_ref=1433.3,
                              dj_c_over_db=500.34,
                              comment='MB INNER                      ')


def test_get_magnum_to_roxie_dct():
    # arrange
    # act
    magnum_to_roxie_dct = strand_def.get_magnum_to_roxie_dct()

    # assert
    magnum_to_roxie_dct_ref = {'name': 'Name', 'd_strand': 'diam.', 'f_cu_nocu': 'cu/sc', 'rrr': 'RRR',
                               'temp_ref': 'Tref', 'b_ref': 'Bref', 'j_c_at_b_ref_t_ref': 'Jc@BrTr',
                               'dj_c_over_db': 'dJc/dB', 'comment': 'Comment'}
    assert magnum_to_roxie_dct_ref == magnum_to_roxie_dct


def test_compute_surface_cu():
    # arrange
    # act
    surface_cu = strand_def.compute_surface_cu()

    # assert
    surface_ref = math.pi * strand_def.d_strand ** 2 / 4
    f_nocu_ref = strand_def.f_cu_nocu / (1 + strand_def.f_cu_nocu)

    assert abs(surface_cu - surface_ref * f_nocu_ref) < 1e-6


def test_compute_surface_nocu():
    # arrange
    # act
    surface_nocu = strand_def.compute_surface_nocu()

    # assert
    surface_ref = math.pi * strand_def.d_strand ** 2 / 4
    f_nocu_ref = 1 / (1 + strand_def.f_cu_nocu)

    assert abs(surface_nocu - surface_ref * f_nocu_ref) < 1e-6


def test_compute_f_nocu():
    # arrange
    # act
    f_nocu = strand_def.compute_f_nocu()

    # assert
    f_nocu_ref = 1 / (1 + strand_def.f_cu_nocu)

    assert abs(f_nocu_ref - f_nocu) < 1e-6


def test_compute_f_cu():
    # arrange
    # act
    f_cu = strand_def.compute_f_cu()

    # assert
    f_cu_ref = strand_def.f_cu_nocu / (1 + strand_def.f_cu_nocu)

    assert abs(f_cu_ref - f_cu) < 1e-6


def test_compute_f_cu_compute_f_nocu():
    # arrange
    # act
    f_cu = strand_def.compute_f_cu()
    f_nocu = strand_def.compute_f_nocu()

    # assert
    assert abs(1 - (f_cu + f_nocu)) < 1e-6


def test_compute_surface():
    # arrange
    # act
    surface = strand_def.compute_surface()

    # assert
    surface_ref = math.pi * strand_def.d_strand ** 2 / 4
    assert abs(surface_ref - surface) < 1e-9
